"""Morphology classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import os
import platform
import logging
import pathlib
import bisect
import numpy
from bluepyopt.ephys.base import BaseEPhys
from bluepyopt.ephys.serializer import DictMixin
from bluepyopt.ephys.acc import arbor, ArbLabel

logger = logging.getLogger(__name__)

# TODO define an addressing scheme


class Morphology(BaseEPhys):

    """Morphology class"""
    pass


class NrnFileMorphology(Morphology, DictMixin):

    """Morphology loaded from a file"""
    SERIALIZED_FIELDS = ('morphology_path', 'do_replace_axon', 'do_set_nseg',
                         'replace_axon_hoc', 'nseg_frequency',
                         'morph_modifiers', 'morph_modifiers_hoc',
                         'morph_modifiers_kwargs')

    def __init__(
            self,
            morphology_path,
            do_replace_axon=False,
            do_set_nseg=True,
            comment='',
            replace_axon_hoc=None,
            nseg_frequency=40,
            morph_modifiers=None,
            morph_modifiers_hoc=None,
            morph_modifiers_kwargs=None):
        """Constructor

        Args:
            morphology_path (str or Path): location of the file describing the
                morphology
            do_replace_axon (bool): Does the axon need to be replaced by an AIS
                stub with default function ?
            replace_axon_hoc (str): Translation in HOC language for the
                'replace_axon' method. This code will 'only' be used when
                calling create_hoc on a cell model. While the model is run in
                python, replace_axon is used instead. Must include
                'proc replace_axon(){ ... }
                If None,the default replace_axon is used
            nseg_frequency (float): frequency of nseg
            do_set_nseg (bool): if True, it will use nseg_frequency
            morph_modifiers (list): list of functions to modify the icell
                with (sim, icell) as arguments
            morph_modifiers_hoc (list): list of hoc strings corresponding
                to morph_modifiers
            morph_modifiers_kwargs (dict): kwargs for morph_modifiers functions
        """
        name = os.path.basename(morphology_path)
        super(NrnFileMorphology, self).__init__(name=name, comment=comment)
        # TODO speed up loading of morphologies from files
        # Path to morphology
        if isinstance(morphology_path, pathlib.Path):
            morphology_path = str(morphology_path)
        self.morphology_path = morphology_path
        self.do_replace_axon = do_replace_axon
        self.do_set_nseg = do_set_nseg
        self.nseg_frequency = nseg_frequency
        self.morph_modifiers = morph_modifiers
        self.morph_modifiers_hoc = morph_modifiers_hoc
        self.morph_modifiers_kwargs = morph_modifiers_kwargs
        if self.morph_modifiers_kwargs is None:
            self.morph_modifiers_kwargs = {}

        if replace_axon_hoc is None:
            self.replace_axon_hoc = self.default_replace_axon_hoc
        else:
            self.replace_axon_hoc = replace_axon_hoc

    def __str__(self):
        """Return string representation"""

        return self.morphology_path

    def instantiate(self, sim=None, icell=None):
        """Load morphology"""

        logger.debug('Loading morphology %s', self.morphology_path)

        if not os.path.exists(self.morphology_path):
            raise IOError(
                'Morphology not found at \'%s\'' %
                self.morphology_path)

        sim.neuron.h.load_file('stdrun.hoc')
        sim.neuron.h.load_file('import3d.hoc')

        extension = self.morphology_path.split('.')[-1]

        if extension.lower() == 'swc':
            imorphology = sim.neuron.h.Import3d_SWC_read()
        elif extension.lower() == 'asc':
            imorphology = sim.neuron.h.Import3d_Neurolucida3()
        else:
            raise ValueError("Unknown filetype: %s" % extension)

        # TODO this is to get rid of stdout print of neuron
        # probably should be more intelligent here, and filter out the
        # lines we don't want

        imorphology.quiet = 1

        if platform.system() == 'Windows':
            sim.neuron.h.hoc_stdout('NUL')
        else:
            sim.neuron.h.hoc_stdout('/dev/null')

        imorphology.input(str(self.morphology_path))
        sim.neuron.h.hoc_stdout()

        morphology_importer = sim.neuron.h.Import3d_GUI(imorphology, 0)

        morphology_importer.instantiate(icell)

        # TODO Set nseg should be called after all the parameters have been
        # set
        # (in case e.g. Ra was changed)
        if self.do_set_nseg:
            self.set_nseg(icell)

        if self.do_replace_axon:
            self.replace_axon(sim=sim, icell=icell)

        if self.morph_modifiers is not None:
            for morph_modifier in self.morph_modifiers:
                morph_modifier(sim=sim, icell=icell,
                               **self.morph_modifiers_kwargs)

    def destroy(self, sim=None):
        """Destroy morphology instantiation"""
        pass

    def set_nseg(self, icell):
        """Set the nseg of every section"""
        for section in icell.all:
            section.nseg = 1 + 2 * int(section.L / self.nseg_frequency)

    @staticmethod
    def replace_axon(sim=None, icell=None):
        """Replace axon"""

        nsec = len([sec for sec in icell.axonal])

        if nsec == 0:
            ais_diams = [1, 1]
        elif nsec == 1:
            ais_diams = [icell.axon[0].diam, icell.axon[0].diam]
        else:
            ais_diams = [icell.axon[0].diam, icell.axon[0].diam]
            # Define origin of distance function
            sim.neuron.h.distance(0, 0.5, sec=icell.soma[0])

            for section in icell.axonal:
                # If distance to soma is larger than 60, store diameter
                if sim.neuron.h.distance(1, 0.5, sec=section) > 60:
                    ais_diams[1] = section.diam
                    break

        for section in icell.axonal:
            sim.neuron.h.delete_section(sec=section)

        # Create new axon array
        sim.neuron.h.execute('create axon[2]', icell)

        for index, section in enumerate(icell.axon):
            section.nseg = 1
            section.L = 30
            section.diam = ais_diams[index]
            icell.axonal.append(sec=section)
            icell.all.append(sec=section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

        logger.debug('Replace axon with AIS')

    default_replace_axon_hoc = \
        '''
proc replace_axon(){ local nSec, D1, D2
  // preserve the number of original axonal sections
  nSec = sec_count(axonal)

  // Try to grab info from original axon
  if(nSec == 0) { //No axon section present
    D1 = D2 = 1
  } else if(nSec == 1) {
    axon[0] D1 = D2 = diam
  } else {
    axon[0] D1 = D2 = diam
    soma distance() //to calculate distance from soma
    forsec axonal{
      //if section is longer than 60um then store diam and exit from loop
      if(distance(0.5) > 60){
        D2 = diam
        break
      }
    }
  }

  // get rid of the old axon
  forsec axonal{
    delete_section()
  }

  create axon[2]

  axon[0] {
    L = 30
    diam = D1
    nseg = 1 + 2*int(L/40)
    all.append()
    axonal.append()
  }
  axon[1] {
    L = 30
    diam = D2
    nseg = 1 + 2*int(L/40)
    all.append()
    axonal.append()
  }
  nSecAxonal = 2
  soma[0] connect axon[0](0), 1
  axon[0] connect axon[1](0), 1
}
        '''


class ArbFileMorphology(Morphology, DictMixin):
    """Arbor morphology utilities"""

    # Arbor morphology tags
    tags = dict(
        soma=1,
        axon=2,
        dend=3,
        apic=4,
        myelin=5
    )

    # Correspondence of BluePyOpt seclists to Arbor region labels
    # (renaming locations according to SWC convention: using
    # 'dend' for basal dendrite, 'apic' for apical dendrite)
    region_labels = dict(
        all=ArbLabel(
            type='region', name='all', s_expr='(all)'),
        somatic=ArbLabel(
            type='region', name='soma', s_expr='(tag %i)' % tags['soma']),
        axonal=ArbLabel(
            type='region', name='axon', s_expr='(tag %i)' % tags['axon']),
        basal=ArbLabel(
            type='region', name='dend', s_expr='(tag %i)' % tags['dend']),
        apical=ArbLabel(
            type='region', name='apic', s_expr='(tag %i)' % tags['apic']),
        myelinated=ArbLabel(
            type='region', name='myelin', s_expr='(tag %i)' % tags['myelin']),
    )

    @staticmethod
    def load(morpho_filename, replace_axon):
        '''Load morphology and optionally perform axon replacement

        Args:
            morpho_filename (str): Path to file with original morphology.
            replace_axon (): Path to/ACC string for morphology to replace
            axon with (if not None).
        '''

        morpho_suffix = pathlib.Path(morpho_filename).suffix

        if morpho_suffix == '.acc':
            morpho = arbor.load_component(morpho_filename).component
        elif morpho_suffix == '.swc':
            morpho = arbor.load_swc_arbor(morpho_filename)
        elif morpho_suffix == '.asc':
            morpho = arbor.load_asc(morpho_filename).morphology
        else:
            raise RuntimeError(
                'Unsupported morphology %s' % morpho_filename +
                ' (only .swc and .asc supported)')

        if replace_axon is not None:
            replacement = arbor.load_component(replace_axon).component
            morpho = ArbFileMorphology.replace_axon(morpho, replacement)

        return morpho

    @staticmethod
    def extract_nrn_seclists(icell, seclists):
        '''Extract section lists from an instantiated cell (axon replacement)

        Args:
            icell (): Instantiated cell model in the NEURON simulator.
            seclists (): List of section lists to extract
            (typically ['axon'] or ['axon', 'myelin']).
        '''
        replace_axon = arbor.segment_tree()
        nrn_seg_to_dist = dict()
        nrn_seg_to_arb_seg = dict()
        for sec in seclists:
            for section in getattr(icell, sec):

                if replace_axon.size == 0:  # root
                    arb_parent_seg = arbor.mnpos
                else:
                    parent_seg = section.parentseg()
                    parent_sec = parent_seg.sec.name()
                    parent_x = parent_seg.x
                    parent_seg_id = bisect.bisect_left(
                        nrn_seg_to_dist[parent_sec],
                        parent_x)
                    arb_parent_seg = \
                        nrn_seg_to_arb_seg[parent_sec][parent_seg_id]

                pts3d = section.psection()['morphology']['pts3d']
                if len(pts3d) == 0:
                    # stylized geometry, must use sim.neuron.h.define_shape()
                    raise ValueError('Before exporting to ACC, embed'
                                     ' stylized geometry in 3d'
                                     ' by instantiating morphology with'
                                     ' cell_model.instantiate_morphology_3d.')

                pts3d = numpy.array(pts3d)
                dist_x = numpy.cumsum(
                    numpy.linalg.norm(
                        pts3d[1:, :3] - pts3d[:-1, :3], axis=1)) /\
                    section.psection()['morphology']['L']
                relative_length_err = abs(1. - dist_x[-1])
                if relative_length_err > 1e-4:
                    logger.warn('pts3d length does not add up to'
                                ' section length, relative error = %s' %
                                relative_length_err)
                    if relative_length_err > 1e-2:
                        raise ValueError('pts3d length inconsistent'
                                         ' with section length, relative'
                                         ' error = %s' %
                                         relative_length_err)

                dist_x[-1] = 1.
                nrn_seg_to_dist[section.name()] = dist_x

                arb_seg_ids = []
                for i in range(1, len(pts3d)):
                    prox = pts3d[i - 1]
                    dist = pts3d[i]
                    arb_parent_seg = replace_axon.append(
                        arb_parent_seg,
                        arbor.mpoint(*prox[:3], 0.5 * prox[3]),
                        arbor.mpoint(*dist[:3], 0.5 * dist[3]),
                        ArbFileMorphology.tags[sec])
                    arb_seg_ids.append(arb_parent_seg)
                # dist, arb_seg_id pairs
                nrn_seg_to_arb_seg[section.name()] = arb_seg_ids

        replace_axon = arbor.morphology(replace_axon)

        return replace_axon

    @staticmethod
    def replace_axon(morphology, replacement=None):
        '''return a morphology with the axon replaced by another morphology

        Args:
            morphology (arbor.morphology): The original Arbor morphology
            replacement (): An Arbor morphology to replace the axon with
        '''

        # Check if tag_roots is available
        if not hasattr(arbor.segment_tree, 'tag_roots'):
            raise NotImplementedError(
                "Need a newer version of Arbor for axon replacement.")

        # Arbor tags
        axon_tag = ArbFileMorphology.tags['axon']

        # prune morphology at axon root
        st = morphology.to_segment_tree()
        axon_roots = st.tag_roots(axon_tag)
        if len(axon_roots) > 1:
            raise ValueError("Axon replacement is only supported for "
                             "morphologies with a single axon root.")
        elif len(axon_roots) == 1:
            axon_root = axon_roots[0]
            logger.debug('Axon replacement: splitting segment tree'
                         ' at segment %d.', axon_root)
            pruned_st, axon_st = st.split_at(axon_root)
            axon_parent = st.parents[axon_root]
        else:
            pruned_st = st
            axon_parent = arbor.mnpos

        # join pruned segment tree and replacement at axon parent
        axon_replacement_st = replacement.to_segment_tree()

        logger.debug('Axon replacement: joining replacement onto'
                     'pruned tree at parent segment %d.', axon_parent)
        joined_st = pruned_st.join_at(
            axon_parent, axon_replacement_st)

        return arbor.morphology(joined_st)



"""Tests for create_acc.py"""

# pylint: disable=W0212

import os
import sys
import pathlib
import re
import json
import tempfile

from bluepyopt.ephys.acc import arbor, ArbLabel
from bluepyopt.ephys.morphologies import ArbFileMorphology
from bluepyopt.ephys.parameterscalers import NrnSegmentSomaDistanceScaler

from . import utils

from bluepyopt import ephys
from bluepyopt.ephys import create_acc
from bluepyopt.ephys.create_acc import (Nrn2ArbParamAdapter,
                                        Nrn2ArbMechGrouper,
                                        ArbNmodlMechFormatter)

import pytest

DEFAULT_ARBOR_REGION_ORDER = [
    ('soma', 1),
    ('axon', 2),
    ('dend', 3),
    ('apic', 4),
    ('myelin', 5)]


testdata_dir = pathlib.Path(__file__).parent.joinpath(
    'testdata')


@pytest.mark.unit
def test_read_templates():
    """Unit test for _read_templates function."""
    template_dir = testdata_dir / 'acc/templates'
    template_filename = "*_template.jinja2"
    templates = create_acc._read_templates(template_dir, template_filename)
    assert templates.keys() == {'label_dict.acc', 'cell.json', 'decor.acc'}

    with pytest.raises(FileNotFoundError):
        create_acc._read_templates("DOES_NOT_EXIST", template_filename)


@pytest.mark.unit
def test_Nrn2ArbParamAdapter_param_name():
    """Test Neuron to Arbor parameter mapping."""
    # Identity
    mech_param_name = "gSKv3_1bar_SKv3_1"
    assert Nrn2ArbParamAdapter._param_name(mech_param_name) \
        == mech_param_name

    # Non-trivial transformation
    global_property_name = "v_init"
    assert Nrn2ArbParamAdapter._param_name(global_property_name) \
        == "membrane-potential"


@pytest.mark.unit
def test_Nrn2ArbParamAdapter_param_value():
    """Test Neuron to Arbor parameter units conversion."""
    # Identity for region parameter
    mech_param = create_acc.Location(name="gSKv3_1bar_SKv3_1", value="1.025")
    assert Nrn2ArbParamAdapter._param_value(mech_param) == "1.025"

    # Non-trivial name transformation, but identical value/units
    global_property = create_acc.Location(name="v_init", value=-65)
    assert Nrn2ArbParamAdapter._param_value(global_property) == "-65"

    # Non-trivial name and value/units transformation
    global_property = create_acc.Location(name="celsius", value=34)
    assert Nrn2ArbParamAdapter._param_value(global_property) == (
        "307.14999999999998")


@pytest.mark.unit
def test_Nrn2ArbParamAdapter_format():
    """Test Neuron to Arbor parameter reformatting."""
    # Constant mechanism parameter
    mech_param = create_acc.Location(name="gSKv3_1bar_SKv3_1", value="1.025")
    mech = "SKv3_1"
    arb_mech_param = create_acc.Location(name="gSKv3_1bar", value="1.025")
    assert (
        Nrn2ArbParamAdapter.format(
            mech_param, mechs=[mech])
        == (mech, arb_mech_param)
    )

    # Non-unique mapping to mechanisms
    with pytest.raises(create_acc.CreateAccException):
        Nrn2ArbParamAdapter.format(
            mech_param, mechs=["SKv3_1", "1"])

    # Global property with non-trivial transformation
    global_property = create_acc.Location(name="celsius", value="0")
    mech = None
    arb_global_property = create_acc.Location(
        name="temperature-kelvin", value="273.14999999999998")
    # Non-trivial name and value/units transformation
    assert Nrn2ArbParamAdapter.format(global_property, []) == \
        (mech, arb_global_property)

    # Inhomogeneuos mechanism parameter
    apical_region = ArbLabel("region", "apic", "(tag 4)")
    param_scaler = NrnSegmentSomaDistanceScaler(
        name='soma-distance-scaler',
        distribution='(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}'
    )

    iexpr_param = create_acc.RangeExpr(
        location=apical_region,
        name="gkbar_hh",
        value="0.025",
        value_scaler=param_scaler
    )
    mech = "hh"
    arb_iexpr_param = create_acc.RangeExpr(
        location=apical_region,
        name="gkbar",
        value="0.025",
        value_scaler=param_scaler,
    )
    assert (
        Nrn2ArbParamAdapter.format(
            iexpr_param, mechs=[mech])
        == (mech, arb_iexpr_param)
    )

    # Point process mechanism parameter
    loc = ephys.locations.ArbLocsetLocation(
        name='somacenter',
        locset='(location 0 0.5)')

    mech = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='ExpSyn',
        locations=[loc])

    mech_loc = ephys.locations.NrnPointProcessLocation(
        'expsyn_loc',
        pprocess_mech=mech)

    point_expr_param = create_acc.PointExpr(
        name="tau", value="10", point_loc=[mech_loc])

    arb_point_expr_param = create_acc.PointExpr(
        name="tau", value="10", point_loc=[mech_loc])
    assert (
        Nrn2ArbParamAdapter.format(
            point_expr_param, mechs=[mech])
        == (mech, arb_point_expr_param)
    )


@pytest.mark.unit
def test_Nrn2ArbMechGrouper_format_params_and_group_by_mech():
    """Test grouping of parameters by mechanism."""
    params = [create_acc.Location(name="gSKv3_1bar_SKv3_1", value="1.025"),
              create_acc.Location(name="ena", value="-30")]
    mechs = ["SKv3_1"]
    local_mechs = Nrn2ArbMechGrouper.\
        _format_params_and_group_by_mech(params, mechs)
    assert local_mechs == \
        {None: [create_acc.Location(name="ion-reversal-potential \"na\"",
                                    value="-30")],
         "SKv3_1": [create_acc.Location(name="gSKv3_1bar", value="1.025")]}


@pytest.mark.unit
def test_Nrn2ArbMechGrouper_process_global():
    """Test adapting global parameters from Neuron to Arbor."""
    params = {"ki": 3, "v_init": -65}
    global_mechs = Nrn2ArbMechGrouper.process_global(params)
    assert global_mechs == {
        None: [create_acc.Location(name="ion-internal-concentration \"k\"",
                                   value="3"),
               create_acc.Location(name="membrane-potential",
                                   value="-65")]}


@pytest.mark.unit
def test_Nrn2ArbMechGrouper_is_global_property():
    """Test adapting local parameters from Neuron to Arbor."""
    all_regions = ArbLabel("region", "all_regions", "(all)")
    param = create_acc.Location(name="axial-resistivity", value="1")
    assert Nrn2ArbMechGrouper._is_global_property(
        all_regions, param) is True

    soma_region = ArbLabel("region", "soma", "(tag 1)")
    assert Nrn2ArbMechGrouper._is_global_property(
        soma_region, param) is False


@pytest.mark.unit
def test_separate_global_properties():
    """Test separating global properties from label-specific mechs."""
    all_regions = ArbLabel("region", "all_regions", "(all)")
    mechs = {None: [create_acc.Location(name="axial-resistivity", value="1")],
             "SKv3_1": [create_acc.Location(name="gSKv3_1bar", value="1.025")]}
    local_mechs, global_properties = \
        Nrn2ArbMechGrouper._separate_global_properties(all_regions, mechs)
    assert local_mechs == {None: [], "SKv3_1": mechs["SKv3_1"]}
    assert global_properties == {None: mechs[None]}


@pytest.mark.unit
def test_Nrn2ArbMechGrouper_process_local():
    """Test adapting local parameters from Neuron to Arbor."""
    all_regions = ArbLabel("region", "all_regions", "(all)")
    soma_region = ArbLabel("region", "soma", "(tag 1)")
    params = [
        (all_regions,
         [create_acc.Location(name="cm", value="100")]),
        (soma_region,
         [create_acc.Location(name="v_init", value="-65"),
          create_acc.Location(name="gSKv3_1bar_SKv3_1", value="1.025")])
    ]
    channels = {all_regions: [], soma_region: ["SKv3_1"]}
    local_mechs, global_properties = \
        Nrn2ArbMechGrouper.process_local(params, channels)
    assert local_mechs.keys() == {all_regions, soma_region}
    assert local_mechs[all_regions] == {None: []}
    assert local_mechs[soma_region] == {
        None: [create_acc.Location(name="membrane-potential", value="-65")],
        "SKv3_1": [create_acc.Location(name="gSKv3_1bar", value="1.025")]
    }
    assert global_properties == {
        None: [create_acc.Location(name="membrane-capacitance", value="1")]}


@pytest.mark.unit
def test_ArbNmodlMechFormatter_load_mech_catalogue_meta():
    """Test loading Arbor built-in mech catalogue metadata."""
    nmodl_formatter = ArbNmodlMechFormatter(None)

    assert isinstance(nmodl_formatter.cats, dict)
    assert nmodl_formatter.cats.keys() == {'BBP', 'default', 'allen'}
    assert "Ca_HVA" in nmodl_formatter.cats['BBP']


@pytest.mark.unit
def test_ArbNmodlMechFormatter_mech_name():
    """Test mechanism name translation."""
    assert ArbNmodlMechFormatter._mech_name("Ca_HVA") == "Ca_HVA"
    assert ArbNmodlMechFormatter._mech_name("ExpSyn") == "expsyn"


@pytest.mark.unit
def test_ArbNmodlMechFormatter_translate_density():
    """Test NMODL GLOBAL parameter handling in mechanism translation."""
    mechs = {
        "hh": [
            create_acc.Location(name="gnabar", value="0.10000000000000001"),
            create_acc.RangeIExpr(
                name="gkbar",
                value="0.029999999999999999",
                scale=(
                    "(add (scalar -0.62109375) (mul (scalar 0.546875) "
                    "(log (add (mul (distance (region \"soma\"))"
                    " (scalar 0.421875) ) (scalar 1.25) ) ) ) )"
                ),
            ),
        ],
        "pas": [
            create_acc.Location(name="e", value="0.25"),
            create_acc.RangeIExpr(
                name="g",
                value="0.029999999999999999",
                scale=(
                    "(add (scalar -0.62109375) (mul (scalar 0.546875) "
                    "(log (add (mul (distance (region \"soma\"))"
                    " (scalar 0.421875) ) (scalar 1.25) ) ) ) )"
                ),
            ),
        ],
    }
    nmodl_formatter = ArbNmodlMechFormatter(None)
    translated_mechs = nmodl_formatter.translate_density(mechs)
    assert translated_mechs.keys() == {"default::hh",
                                       "default::pas/e=0.25"}
    assert translated_mechs["default::hh"] == mechs["hh"]
    assert translated_mechs["default::pas/e=0.25"] == mechs["pas"][1:]


@pytest.mark.unit
def test_arb_populate_label_dict():
    """Unit test for _populate_label_dict."""
    local_mechs = {ArbLabel("region", "all", "(all)"): {}}
    local_scaled_mechs = {
        ArbLabel("region", "first_branch", "(branch 0)"): {}}
    pprocess_mechs = {}

    label_dict = create_acc._arb_populate_label_dict(local_mechs,
                                                     local_scaled_mechs,
                                                     pprocess_mechs)
    assert label_dict.keys() == {"all", "first_branch"}

    with pytest.raises(create_acc.CreateAccException):
        other_pprocess_mechs = {
            ArbLabel("region", "first_branch", "(branch 1)"): {}}
        create_acc._arb_populate_label_dict(local_mechs,
                                            local_scaled_mechs,
                                            other_pprocess_mechs)


@pytest.mark.unit
def test_create_acc():
    """ephys.create_acc: Test create_acc"""
    mech = utils.make_mech()
    parameters = utils.make_parameters()

    acc = create_acc.create_acc([mech, ], parameters,
                                morphology='CCell.swc',
                                template_name='CCell')

    ref_dir = testdata_dir / 'acc/CCell'
    cell_json = "CCell.json"
    decor_acc = "CCell_decor.acc"
    label_dict_acc = "CCell_label_dict.acc"

    # Testing keys
    assert cell_json in acc
    cell_json_dict = json.loads(acc[cell_json])
    assert 'cell_model_name' in cell_json_dict
    assert 'produced_by' in cell_json_dict
    assert 'morphology' in cell_json_dict
    assert 'label_dict' in cell_json_dict
    assert 'decor' in cell_json_dict
    # Testing values
    with open(ref_dir / cell_json) as f:
        ref_cell_json = json.load(f)
    for k in ref_cell_json:
        if k != 'produced_by':
            assert ref_cell_json[k] == cell_json_dict[k]

    # Testing building blocks
    assert decor_acc in acc
    assert acc[decor_acc].startswith('(arbor-component')
    assert '(decor' in acc[decor_acc]
    # Testing values
    with open(ref_dir / decor_acc) as f:
        ref_decor = f.read()
    assert ref_decor == acc[decor_acc]  # decor data not exposed in Python

    # Testing building blocks
    assert label_dict_acc in acc
    assert acc[label_dict_acc].startswith('(arbor-component')
    assert '(label-dict' in acc[label_dict_acc]
    matches = re.findall(r'\(region-def "(?P<loc>\w+)" \(tag (?P<tag>\d+)\)\)',
                         acc[label_dict_acc])
    for pos, loc_tag in enumerate(DEFAULT_ARBOR_REGION_ORDER):
        assert matches[pos][0] == loc_tag[0]
        assert matches[pos][1] == str(loc_tag[1])
    # Testing values
    ref_labels = arbor.load_component(
        ref_dir / label_dict_acc).component
    with tempfile.TemporaryDirectory() as test_dir:
        test_labels_filename = pathlib.Path(test_dir).joinpath(label_dict_acc)
        with open(test_labels_filename, 'w') as f:
            f.write(acc[label_dict_acc])
        test_labels = arbor.load_component(test_labels_filename).component
    assert dict(ref_labels.items()) == dict(test_labels.items())


@pytest.mark.unit
def test_create_acc_filename():
    """ephys.create_acc: Test create_acc template_filename"""
    mech = utils.make_mech()
    parameters = utils.make_parameters()
    custom_param_val = str(__file__)

    acc = create_acc.create_acc(
        [mech, ],
        parameters, morphology='CCell.asc',
        template_name='CCell',
        template_filename='acc/templates/*_template.jinja2',
        template_dir=testdata_dir,
        custom_jinja_params={
            'custom_param': custom_param_val})
    cell_json = "CCell_cell.json"
    decor_acc = "CCell_decor.acc"
    label_dict_acc = "CCell_label_dict.acc"

    assert cell_json in acc
    cell_json_dict = json.loads(acc[cell_json])
    assert 'cell_model_name' in cell_json_dict
    assert 'produced_by' in cell_json_dict
    assert 'morphology' in cell_json_dict
    assert 'label_dict' in cell_json_dict
    assert 'decor' in cell_json_dict

    assert decor_acc in acc
    assert acc[decor_acc].startswith('(arbor-component')
    assert '(decor' in acc[decor_acc]

    assert label_dict_acc in acc
    assert acc[label_dict_acc].startswith('(arbor-component')
    assert '(label-dict' in acc[label_dict_acc]
    matches = re.findall(r'\(region-def "(?P<loc>\w+)" \(tag (?P<tag>\d+)\)\)',
                         acc[label_dict_acc])
    for pos, loc_tag in enumerate(DEFAULT_ARBOR_REGION_ORDER):
        assert matches[pos][0] == loc_tag[0]
        assert matches[pos][1] == str(loc_tag[1])

    assert '(meta-data (info "test-decor"))' in acc[decor_acc]
    assert '(meta-data (info "test-label-dict"))' in acc[label_dict_acc]
    assert custom_param_val in cell_json_dict['produced_by']


@pytest.mark.unit
def test_create_acc_replace_axon():
    """ephys.create_acc: Test create_acc with axon replacement"""
    mech = utils.make_mech()
    parameters = utils.make_parameters()

    replace_axon_st = arbor.segment_tree()
    latest_seg = arbor.mnpos

    for prox_x, dist_x in [(5, 35), (35, 65)]:
        latest_seg = replace_axon_st.append(
            latest_seg,
            arbor.mpoint(prox_x, 0, 0, 0.5),
            arbor.mpoint(dist_x, 0, 0, 0.5),
            ArbFileMorphology.tags['axon']
        )

    replace_axon = arbor.morphology(replace_axon_st)

    try:
        acc = create_acc.create_acc([mech, ], parameters,
                                    morphology_dir=testdata_dir,
                                    morphology='simple.swc',
                                    template_name='CCell',
                                    replace_axon=replace_axon)
    except Exception as e:  # fail with an older Arbor version
        assert isinstance(e, NotImplementedError)
        assert len(e.args) == 1 and e.args[0] == \
            "Need a newer version of Arbor for axon replacement."
        return

    cell_json = "CCell.json"
    cell_json_dict = json.loads(acc[cell_json])
    assert 'replace_axon' in cell_json_dict['morphology']

    with open(testdata_dir / 'acc/CCell/simple_axon_replacement.acc') as f:
        replace_axon_ref = f.read()

    assert acc[cell_json_dict['morphology']['replace_axon']] == \
        replace_axon_ref


def make_cell(replace_axon):
    morph_filename = testdata_dir / 'simple_ax2.swc'
    morph = ephys.morphologies.NrnFileMorphology(morph_filename,
                                                 do_replace_axon=replace_axon)
    somatic_loc = ephys.locations.NrnSeclistLocation(
        'somatic', seclist_name='somatic')
    mechs = [ephys.mechanisms.NrnMODMechanism(
        name='hh', suffix='hh', locations=[somatic_loc])]
    gkbar_hh_scaler = '(-0.62109375 + 0.546875*math.log(' \
                      '({distance})*0.421875 + 1.25))*{value}'
    params = [
        ephys.parameters.NrnSectionParameter(
            name='gnabar_hh',
            param_name='gnabar_hh',
            locations=[somatic_loc]),
        ephys.parameters.NrnRangeParameter(
            name='gkbar_hh',
            param_name='gkbar_hh',
            value_scaler=ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
                distribution=gkbar_hh_scaler),
            locations=[somatic_loc])]
    return ephys.models.CellModel(
        'simple_ax2',
        morph=morph,
        mechs=mechs,
        params=params)


def run_short_sim(cable_cell):
    # Create cell model
    arb_cell_model = arbor.single_cell_model(cable_cell)
    arb_cell_model.properties.catalogue = arbor.catalogue()
    arb_cell_model.properties.catalogue.extend(
        arbor.default_catalogue(), "default::")
    arb_cell_model.properties.catalogue.extend(
        arbor.bbp_catalogue(), "BBP::")

    # Run a very short simulation to test mechanism instantiation
    arb_cell_model.run(tfinal=0.1)


@pytest.mark.unit
def test_cell_model_write_and_read_acc():
    """ephys.create_acc: Test write_acc and read_acc w/o axon replacement"""
    cell = make_cell(replace_axon=False)
    param_values = {'gnabar_hh': 0.1,
                    'gkbar_hh': 0.03}

    with tempfile.TemporaryDirectory() as acc_dir:
        cell.write_acc(acc_dir, param_values)
        cell_json, arb_morph, arb_decor, arb_labels = \
            create_acc.read_acc(
                pathlib.Path(acc_dir).joinpath(cell.name + '.json'))
    assert 'replace_axon' not in cell_json['morphology']

    cable_cell = arbor.cable_cell(morphology=arb_morph,
                                  decor=arb_decor,
                                  labels=arb_labels)
    assert isinstance(cable_cell, arbor.cable_cell)
    assert len(cable_cell.cables('"soma"')) == 1
    assert len(cable_cell.cables('"axon"')) == 1
    assert len(arb_morph.branch_segments(
        cable_cell.cables('"soma"')[0].branch)) == 5
    assert len(arb_morph.branch_segments(
        cable_cell.cables('"axon"')[0].branch)) == 5

    run_short_sim(cable_cell)


@pytest.mark.unit
def test_cell_model_write_and_read_acc_replace_axon():
    """ephys.create_acc: Test write_acc and read_acc w/ axon replacement"""
    cell = make_cell(replace_axon=True)
    param_values = {'gnabar_hh': 0.1,
                    'gkbar_hh': 0.03}

    with tempfile.TemporaryDirectory() as acc_dir:
        try:
            nrn_sim = ephys.simulators.NrnSimulator()
            cell.write_acc(acc_dir, param_values,
                           sim=nrn_sim)
        except Exception as e:  # fail with an older Arbor version
            assert isinstance(e, NotImplementedError)
            assert len(e.args) == 1 and e.args[0] == \
                "Need a newer version of Arbor for axon replacement."
            return

        # Axon replacement implemented in installed Arbor version
        cell_json, arb_morph, arb_decor, arb_labels = \
            create_acc.read_acc(
                pathlib.Path(acc_dir).joinpath(cell.name + '.json'))

    assert 'replace_axon' in cell_json['morphology']
    cable_cell = arbor.cable_cell(morphology=arb_morph,
                                  decor=arb_decor,
                                  labels=arb_labels)
    assert isinstance(cable_cell, arbor.cable_cell)
    assert len(cable_cell.cables('"soma"')) == 1
    assert len(cable_cell.cables('"axon"')) == 1
    assert len(arb_morph.branch_segments(
        cable_cell.cables('"soma"')[0].branch)) == 6
    assert len(arb_morph.branch_segments(
        cable_cell.cables('"axon"')[0].branch)) == 6
    assert cable_cell.cables('"soma"')[0].prox == 0.
    assert abs(cable_cell.cables('"soma"')[0].dist -
               cable_cell.cables('"axon"')[0].prox) < 1e-6
    assert cable_cell.cables('"axon"')[0].dist == 1.

    run_short_sim(cable_cell)


@pytest.mark.unit
def test_cell_model_create_acc_replace_axon_without_instantiate():
    """ephys.create_acc: Test write_acc and read_acc w/ axon replacement"""
    cell = make_cell(replace_axon=True)
    param_values = {'gnabar_hh': 0.1,
                    'gkbar_hh': 0.03}

    with pytest.raises(ValueError,
                       match='Need an instance of NrnSimulator in sim'
                             ' to instantiate morphology in order to'
                             ' create JSON/ACC-description with'
                             ' axon replacement.'):
        cell.create_acc(param_values)


def check_acc_dir(test_dir, ref_dir):
    assert os.listdir(ref_dir) == os.listdir(test_dir)

    ref_dir_ver_suffix = '_py' + ''.join(sys.version.split('.')[:2])
    ref_dir_ver = ref_dir.parent / (ref_dir.name + ref_dir_ver_suffix)

    for file in os.listdir(ref_dir):

        if (ref_dir_ver / file).exists():
            ref_dir_file = ref_dir_ver
        else:
            ref_dir_file = ref_dir

        if file.endswith('.json'):
            with open(os.path.join(test_dir, file)) as f:
                cell_json_dict = json.load(f)
            with open(ref_dir_file / file) as f:
                ref_cell_json = json.load(f)
            for k in ref_cell_json:
                if k != 'produced_by':
                    assert ref_cell_json[k] == cell_json_dict[k]
        else:
            with open(os.path.join(test_dir, file)) as f:
                test_file = f.read()
            with open(ref_dir_file / file) as f:
                ref_file = f.read()
            assert ref_file == test_file


@pytest.mark.unit
def test_write_acc_simple():
    SIMPLECELL_PATH = str((pathlib.Path(__file__).parent /
                          '../../../examples/simplecell').resolve())
    sys.path.insert(0, SIMPLECELL_PATH)
    ref_dir = (testdata_dir / 'acc/simplecell').resolve()
    old_cwd = os.getcwd()
    try:
        os.chdir(SIMPLECELL_PATH)
        import simplecell_model
        param_values = {
            'gnabar_hh': 0.10299326453483033,
            'gkbar_hh': 0.027124836082684685
        }

        cell = simplecell_model.create(do_replace_axon=True)
        nrn_sim = ephys.simulators.NrnSimulator()
        cell.instantiate_morphology_3d(nrn_sim)

        with tempfile.TemporaryDirectory() as test_dir:
            cell.write_acc(test_dir,
                           param_values,
                           # ext_catalogues=ext_catalogues,
                           create_mod_morph=True)

            check_acc_dir(test_dir, ref_dir)
    except Exception as e:  # fail with an older Arbor version
        assert isinstance(e, NotImplementedError)
        assert len(e.args) == 1 and e.args[0] == \
            "Need a newer version of Arbor for axon replacement."
    finally:
        cell.destroy(nrn_sim)
        os.chdir(old_cwd)
        sys.path.pop(0)


@pytest.mark.unit
def test_write_acc_l5pc():
    L5PC_PATH = str((pathlib.Path(__file__).parent /
                    '../../../examples/l5pc').resolve())
    sys.path.insert(0, L5PC_PATH)
    ref_dir = (testdata_dir / 'acc/l5pc').resolve()
    old_cwd = os.getcwd()
    try:
        import l5pc_model
        param_values = {
            'gNaTs2_tbar_NaTs2_t.apical': 0.026145,
            'gSKv3_1bar_SKv3_1.apical': 0.004226,
            'gImbar_Im.apical': 0.000143,
            'gNaTa_tbar_NaTa_t.axonal': 3.137968,
            'gK_Tstbar_K_Tst.axonal': 0.089259,
            'gamma_CaDynamics_E2.axonal': 0.002910,
            'gNap_Et2bar_Nap_Et2.axonal': 0.006827,
            'gSK_E2bar_SK_E2.axonal': 0.007104,
            'gCa_HVAbar_Ca_HVA.axonal': 0.000990,
            'gK_Pstbar_K_Pst.axonal': 0.973538,
            'gSKv3_1bar_SKv3_1.axonal': 1.021945,
            'decay_CaDynamics_E2.axonal': 287.198731,
            'gCa_LVAstbar_Ca_LVAst.axonal': 0.008752,
            'gamma_CaDynamics_E2.somatic': 0.000609,
            'gSKv3_1bar_SKv3_1.somatic': 0.303472,
            'gSK_E2bar_SK_E2.somatic': 0.008407,
            'gCa_HVAbar_Ca_HVA.somatic': 0.000994,
            'gNaTs2_tbar_NaTs2_t.somatic': 0.983955,
            'decay_CaDynamics_E2.somatic': 210.485284,
            'gCa_LVAstbar_Ca_LVAst.somatic': 0.000333,
        }

        cell = l5pc_model.create(do_replace_axon=True)
        nrn_sim = ephys.simulators.NrnSimulator()
        cell.instantiate_morphology_3d(nrn_sim)

        with tempfile.TemporaryDirectory() as test_dir:
            cell.write_acc(test_dir,
                           param_values,
                           # ext_catalogues=ext_catalogues,
                           create_mod_morph=True)

            check_acc_dir(test_dir, ref_dir)
    except Exception as e:  # fail with an older Arbor version
        assert isinstance(e, NotImplementedError)
        assert len(e.args) == 1 and e.args[0] == \
            "Need a newer version of Arbor for axon replacement."
    finally:
        cell.destroy(nrn_sim)
        os.chdir(old_cwd)
        sys.path.pop(0)


@pytest.mark.unit
def test_write_acc_expsyn():
    EXPSYN_PATH = str((pathlib.Path(__file__).parent /
                      '../../../examples/expsyn').resolve())
    sys.path.insert(0, EXPSYN_PATH)
    ref_dir = (testdata_dir / 'acc/expsyn').resolve()
    old_cwd = os.getcwd()
    try:
        import expsyn
        param_values = {'expsyn_tau': 10.0}

        cell = expsyn.create_model(sim='arb', do_replace_axon=False)

        with tempfile.TemporaryDirectory() as test_dir:
            cell.write_acc(test_dir,
                           param_values,
                           # ext_catalogues=ext_catalogues,
                           create_mod_morph=True)

            check_acc_dir(test_dir, ref_dir)
    finally:
        os.chdir(old_cwd)
        sys.path.pop(0)



"""Extra features functions"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy as np

all_1D_features = [
    "peak_to_valley",
    "halfwidth",
    "peak_trough_ratio",
    "repolarization_slope",
    "recovery_slope",
    "neg_peak_relative",
    "pos_peak_relative",
    "neg_peak_diff",
    "pos_peak_diff",
    "neg_image",
    "pos_image",
]


def calculate_features(
    waveforms,
    sampling_frequency,
    upsample=None,
    feature_names=None,
    recovery_slope_window=0.7
):
    """Calculate features for all waveforms

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    sampling_frequency  : float
        rate at which the waveforms are sampled (Hz)
    feature_names : list or None (if None, compute all)
        features to compute
    recovery_slope_window : float
        windowlength in ms after peak wherein recovery slope is computed

    Returns
    -------
    metrics : dict  (num_waveforms x num_metrics)
        Dictionary with computed metrics. Keys are the metric names, values
            are the computed features

    """
    metrics = dict()

    if feature_names is None:
        feature_names = all_1D_features
    else:
        for name in feature_names:
            assert name in all_1D_features, f"{name} not in {all_1D_features}"

    if upsample is not None:
        assert upsample > 0
        waveforms = _upsample_wf(waveforms, int(upsample))
        sampling_frequency = upsample * sampling_frequency

    if "peak_to_valley" in feature_names:
        metrics["peak_to_valley"] = peak_to_valley(
            waveforms=waveforms, sampling_frequency=sampling_frequency
        )
    if "peak_trough_ratio" in feature_names:
        metrics["peak_trough_ratio"] = peak_trough_ratio(waveforms=waveforms)

    if "halfwidth" in feature_names:
        metrics["halfwidth"] = halfwidth(
            waveforms=waveforms, sampling_frequency=sampling_frequency
        )

    if "repolarization_slope" in feature_names:
        metrics["repolarization_slope"] = repolarization_slope(
            waveforms=waveforms,
            sampling_frequency=sampling_frequency,
        )

    if "recovery_slope" in feature_names:
        metrics["recovery_slope"] = recovery_slope(
            waveforms=waveforms,
            sampling_frequency=sampling_frequency,
            window=recovery_slope_window,
        )

    if "neg_peak_diff" in feature_names:
        metrics["neg_peak_diff"] = peak_time_diff(
            waveforms=waveforms, fs=sampling_frequency, sign="negative"
        )

    if "pos_peak_diff" in feature_names:
        metrics["pos_peak_diff"] = peak_time_diff(
            waveforms=waveforms, fs=sampling_frequency, sign="positive"
        )

    if "neg_peak_relative" in feature_names:
        metrics["neg_peak_relative"] = relative_amplitude(
            waveforms=waveforms, sign="negative"
        )

    if "pos_peak_relative" in feature_names:
        metrics["pos_peak_relative"] = relative_amplitude(
            waveforms=waveforms, sign="positive"
        )

    if "neg_image" in feature_names:
        metrics["neg_image"] = peak_image(waveforms=waveforms, sign="negative")

    if "pos_image" in feature_names:
        metrics["pos_image"] = peak_image(waveforms=waveforms, sign="positive")

    return metrics


def peak_to_valley(waveforms, sampling_frequency):
    """
    Time between trough and peak. If the peak precedes the trough,
    peak_to_valley is negative.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute feature for
    sampling_frequency  : float
        rate at which the waveforms are sampled (Hz)

    Returns
    -------
    np.ndarray (num_waveforms)
        peak_to_valley in seconds

    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    ptv = (peak_idx - trough_idx) * (1 / sampling_frequency)
    ptv[ptv == 0] = np.nan
    return ptv


def peak_trough_ratio(waveforms):
    """
    Normalized ratio peak height and trough depth

    Assumes baseline is 0

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute feature for

    Returns
    -------

    np.ndarray (num_waveforms)
        Peak to trough ratio

    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    ptratio = np.empty(trough_idx.shape[0])
    ptratio[:] = np.nan
    for i in range(waveforms.shape[0]):
        if peak_idx[i] == 0 and trough_idx[i] == 0:
            continue
        ptratio[i] = np.abs(waveforms[i, peak_idx[i]] /
                            waveforms[i, trough_idx[i]])

    return ptratio


def halfwidth(waveforms, sampling_frequency, return_idx=False):
    """
    Width of waveform at its half of amplitude.
    If the peak precedes the trough, halfwidth is negative.

    Computes the width of the waveform peak at half it's height

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    sampling_frequency  : float
        rate at which the waveforms are sampled (Hz)
    return_idx : bool
        if true, also returns index of threshold crossing before and
        index of threshold crossing after peak

    Returns
    -------

    np.ndarray or (np.ndarray, np.ndarray, np.ndarray)
        Halfwidth of the waveforms or (Halfwidth of the waveforms,
        index_cross_pre_peak, index_cross_post_peak)

    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    hw = np.empty(waveforms.shape[0])
    hw[:] = np.nan
    cross_pre_pk = np.empty(waveforms.shape[0], dtype=int)
    cross_post_pk = np.empty(waveforms.shape[0], dtype=int)

    for i in range(waveforms.shape[0]):
        if peak_idx[i] >= trough_idx[i]:
            trough_val = waveforms[i, trough_idx[i]]
            threshold = (
                0.5 * trough_val
            )  # threshold is half of peak heigth (assuming baseline is 0)

            cpre_idx = np.where(waveforms[i, :trough_idx[i]] < threshold)[0]
            cpost_idx = np.where(waveforms[i, trough_idx[i]:] < threshold)[0]

            if len(cpre_idx) == 0 or len(cpost_idx) == 0:
                continue

            cross_pre_pk[i] = (
                cpre_idx[0] - 1
            )  # last occurence of waveform lower than thr, before peak
            cross_post_pk[i] = (
                cpost_idx[-1] + 1 + trough_idx[i]
            )  # first occurence of waveform lower than peak, after peak

            hw[i] = (cross_post_pk[i] - cross_pre_pk[i]) * (
                1 / sampling_frequency
            )  # + peak_idx[i]
        else:
            peak_val = waveforms[i, peak_idx[i]]
            threshold = (
                0.5 * peak_val
            )  # threshold is half of peak heigth (assuming baseline is 0)

            cpre_idx = np.where(waveforms[i, :peak_idx[i]] > threshold)[0]
            cpost_idx = np.where(waveforms[i, peak_idx[i]:] > threshold)[0]

            if len(cpre_idx) == 0 or len(cpost_idx) == 0:
                continue

            cross_pre_pk[i] = (
                cpre_idx[0] - 1
            )  # last occurence of waveform lower than thr, before peak
            cross_post_pk[i] = (
                cpost_idx[-1] + 1 + trough_idx[i]
            )  # first occurence of waveform lower than peak, after peak

            hw[i] = -(cross_post_pk[i] - cross_pre_pk[i]) * (
                1 / sampling_frequency
            )  # + peak_idx[i]

    if not return_idx:
        return hw
    else:
        return hw, cross_pre_pk, cross_post_pk


def repolarization_slope(waveforms, sampling_frequency, return_idx=False):
    """
    Return slope of repolarization period between trough and baseline

    After reaching its maxumum polarization, the neuron potential will
    recover. The repolarization slope is defined as the dV/dT of the action
    potential between trough and baseline.

    Optionally the function returns also the indices per waveform where the
    potential crosses baseline.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    sampling_frequency  : float
        rate at which the waveforms are sampled (Hz)
    return_idx : bool
        if true, also returns index of threshold crossing before and
        index of threshold crossing after peak

    Returns
    -------

    np.ndarray or (np.ndarray, np.ndarray)
        Repolarization slope of the waveforms or (Repolarization slope of the
        waveforms, return to base index)
    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)

    rslope = np.empty(waveforms.shape[0])
    rslope[:] = np.nan
    return_to_base_idx = np.empty(waveforms.shape[0], dtype=np.int_)
    return_to_base_idx[:] = 0

    time = np.arange(0, waveforms.shape[1]) * (1 / sampling_frequency)  # in s
    for i in range(waveforms.shape[0]):
        if trough_idx[i] == 0:
            continue

        rtrn_idx = np.where(waveforms[i, trough_idx[i]:] >= 0)[0]
        if len(rtrn_idx) == 0:
            continue

        return_to_base_idx[i] = (
            rtrn_idx[0] + trough_idx[i]
        )  # first time after  trough, where waveform is at baseline

        if return_to_base_idx[i] - trough_idx[i] < 3:
            continue
        slope = _get_slope(
            time[trough_idx[i]:return_to_base_idx[i]],
            waveforms[i, trough_idx[i]:return_to_base_idx[i]]
        )
        rslope[i] = slope[0]

    if not return_idx:
        return rslope
    else:
        return rslope, return_to_base_idx


def recovery_slope(waveforms, sampling_frequency, window):
    """
    Return the recovery slope of input waveforms. After repolarization,
    the neuron hyperpolarizes until it peaks. The recovery slope is the
    slope of the action potential after the peak, returning to the baseline
    in dV/dT. The slope is computed within a user-defined window after
    the peak.

    Takes a numpy array of waveforms and returns an array with
    recovery slopes per waveform.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    sampling_frequency  : float
        rate at which the waveforms are sampled (Hz)
    window : float
        length after peak wherein to compute recovery slope (ms)


    Returns
    -------
    np.ndarray
        Recovery slope of the waveforms
    """
    _, peak_idx = _get_trough_and_peak_idx(waveforms)
    rslope = np.empty(waveforms.shape[0])
    rslope[:] = np.nan

    time = np.arange(0, waveforms.shape[1]) * (1 / sampling_frequency)  # in s

    for i in range(waveforms.shape[0]):
        if peak_idx[i] in [0, waveforms.shape[1]]:
            continue
        max_idx = int(peak_idx[i] + ((window / 1000) * sampling_frequency))
        max_idx = np.min([max_idx, waveforms.shape[1]])

        if len(time[peak_idx[i]:max_idx]) < 3:
            continue
        slope = _get_slope(
            time[peak_idx[i]:max_idx], waveforms[i, peak_idx[i]:max_idx]
        )
        rslope[i] = slope[0]

    return rslope


def peak_image(waveforms, sign="negative"):
    """
    Normalized amplitude at the time of minimum or maximum peak.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    sign : str
        "pos" | "neg"

    Returns
    -------
    np.ndarray
        Peak images for the waveforms

    """
    assert len(waveforms) > 1

    if sign == "negative":
        funarg = np.argmin
        fun = np.min
    else:
        funarg = np.argmax
        fun = np.max

    peak_channel, peak_time = np.unravel_index(
        funarg(waveforms), waveforms.shape
    )
    relative_peaks = waveforms[:, peak_time] / fun(waveforms[peak_channel])

    return relative_peaks


def relative_amplitude(waveforms, sign="negative"):
    """
    Normalized amplitude with respect to channel with largest amplitude.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    fs : float
        Sampling rate in Hz
    sign : str
        "positive" | "negative"

    Returns
    -------
    np.ndarray
        Relative amplitudes for the waveforms

    """
    assert len(waveforms) > 1

    if sign == "negative":
        fun = np.min
    else:
        fun = np.max

    peak_amp = np.abs(fun(waveforms))
    relative_peaks = np.abs(fun(waveforms, 1)) / peak_amp

    return relative_peaks


def peak_time_diff(waveforms, fs, sign="negative"):
    """
    Peak time differences with respect to channel with largest amplitude.

    Parameters
    ----------
    waveforms  : numpy.ndarray (num_waveforms x num_samples)
        waveforms to compute features for
    fs : float
        Sampling rate in Hz
    sign : str
        "positive" | "negative"

    Returns
    -------
    np.ndarray
        Peak time differences for the waveforms

    """
    assert len(waveforms) > 1

    if sign == "negative":
        argfun = np.argmin
    else:
        argfun = np.argmax

    peak_chan = np.unravel_index(argfun(waveforms), waveforms.shape)[0]
    peak_time = argfun(waveforms[peak_chan])
    relative_peak_times = (argfun(waveforms, 1) - peak_time) / fs

    return relative_peak_times


def _get_slope(x, y):
    """
    Retrun the slope of x and y data, using scipy.signal.linregress
    """
    from scipy.stats import linregress

    slope = linregress(x, y)
    return slope


def _get_trough_and_peak_idx(waveform, after_max_trough=False):
    """
    Return the indices into the input waveforms of the detected troughs
    (minimum of waveform) and peaks (maximum of waveform, after trough).

    Assumes negative troughs and positive peaks

    Returns 0 if not detected
    """
    if after_max_trough:
        max_through_idx = np.unravel_index(
            np.argmin(waveform),
            waveform.shape)[1]
        trough_idx = (
            np.argmin(waveform[:, max_through_idx:], axis=1) + max_through_idx
        )
        peak_idx = (
            np.argmax(waveform[:, max_through_idx:], axis=1) + max_through_idx
        )
    else:
        trough_idx = np.argmin(waveform, axis=1)
        peak_idx = np.argmax(waveform, axis=1)

    return trough_idx, peak_idx


def _upsample_wf(waveforms, upsample):
    from scipy.signal import resample_poly

    ndim = len(waveforms.shape)
    waveforms_up = resample_poly(waveforms, up=upsample, down=1, axis=ndim - 1)

    return waveforms_up



"""Tests for create_hoc.py"""

# pylint: disable=W0212

import os

from bluepyopt.ephys.acc import ArbLabel
from bluepyopt.ephys.parameterscalers import NrnSegmentSomaDistanceScaler

from . import utils
from bluepyopt.ephys import create_acc, create_hoc


import pytest

DEFAULT_LOCATION_ORDER = [
    'all',
    'apical',
    'axonal',
    'basal',
    'somatic',
    'myelinated']


@pytest.mark.unit
def test_generate_channels_by_location():
    """ephys.create_hoc: Test generate_channels_by_location"""
    mech = utils.make_mech()
    public_res = create_hoc.generate_channels_by_location(
        [mech], DEFAULT_LOCATION_ORDER,
    )
    private_res = create_hoc._generate_channels_by_location(
        [mech], DEFAULT_LOCATION_ORDER, create_hoc._loc_desc
    )
    assert public_res == private_res


@pytest.mark.unit
def test__generate_channels_by_location():
    """ephys.create_hoc: Test _generate_channels_by_location"""
    mech = utils.make_mech()
    channels, point_channels = create_hoc._generate_channels_by_location(
        [mech, ], DEFAULT_LOCATION_ORDER, create_hoc._loc_desc)

    assert len(channels['apical']) == 1
    assert len(channels['basal']) == 1

    assert channels['apical'] == ['Ih']
    assert channels['basal'] == ['Ih']

    for loc in point_channels:
        assert len(point_channels[loc]) == 0


@pytest.mark.unit
def test_generate_parameters():
    """ephys.create_hoc: Test generate_parameters"""
    parameters = utils.make_parameters()

    assert create_hoc.generate_parameters(parameters) == \
        create_hoc._generate_parameters(parameters,
                                        DEFAULT_LOCATION_ORDER,
                                        create_hoc._loc_desc)


@pytest.mark.unit
def test__generate_parameters():
    """ephys.create_hoc: Test _generate_parameters"""
    parameters = utils.make_parameters()

    global_params, section_params, range_params, \
        pprocess_params, location_order = \
        create_hoc._generate_parameters(parameters,
                                        DEFAULT_LOCATION_ORDER,
                                        create_hoc._loc_desc)

    assert global_params == {'gSKv3_1bar_SKv3_1': 65}
    assert len(section_params[1]) == 2
    assert len(section_params[4]) == 2
    assert section_params[4][0] == 'somatic'
    assert len(section_params[4][1]) == 2
    assert range_params == []
    for loc, pparams in pprocess_params:
        assert loc in DEFAULT_LOCATION_ORDER
        assert len(pparams) == 0
    assert location_order == DEFAULT_LOCATION_ORDER


@pytest.mark.unit
def test_create_hoc():
    """ephys.create_hoc: Test create_hoc"""
    mech = utils.make_mech()
    parameters = utils.make_parameters()

    hoc = create_hoc.create_hoc([mech, ], parameters, template_name='CCell')
    assert 'load_file' in hoc
    assert 'CCell' in hoc
    assert 'begintemplate' in hoc
    assert 'endtemplate' in hoc


@pytest.mark.unit
def test_create_hoc_filename():
    """ephys.create_hoc: Test create_hoc template_filename"""
    mech = utils.make_mech()
    parameters = utils.make_parameters()
    custom_param_val = 'printf("Hello world!")'

    hoc = create_hoc.create_hoc([mech, ],
                                parameters, template_name='CCell',
                                template_filename='test.jinja2',
                                template_dir=os.path.join(
                                    os.path.dirname(__file__),
                                    'testdata'),
                                custom_jinja_params={
                                    'custom_param': custom_param_val})
    assert 'load_file' in hoc
    assert 'CCell' in hoc
    assert 'begintemplate' in hoc
    assert 'endtemplate' in hoc
    assert 'Test template' in hoc
    assert custom_param_val in hoc


@pytest.mark.unit
def test_generate_reinitrng():
    """ephys.create_hoc: Test generate_reinitrng"""
    mech = utils.make_mech()
    re_init_rng = create_hoc.generate_reinitrng([mech])
    assert 'func hash_str() {localobj sf strdef right' in re_init_rng
    assert ' hash = (hash * 31 + char_int) % (2 ^ 31 - 1)' in re_init_rng


@pytest.mark.unit
def test_range_exprs_to_hoc():
    """ephys.create_hoc: Test range_exprs_to_hoc"""
    apical_region = ArbLabel("region", "apic", "(tag 4)")
    param_scaler = NrnSegmentSomaDistanceScaler(
        name='soma-distance-scaler',
        distribution='(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}'
    )

    range_expr = create_acc.RangeExpr(
        location=apical_region,
        name="gkbar_hh",
        value=0.025,
        value_scaler=param_scaler
    )

    hoc = create_hoc.range_exprs_to_hoc([range_expr])
    assert hoc[0].param_name == 'gkbar_hh'
    val_gt = '(-0.8696 + 2.087*exp((%.17g)*0.0031))*0.025000000000000001'
    assert hoc[0].value == val_gt



"""bluepyopt.optimisations tests"""


import bluepyopt.optimisations
import bluepyopt.ephys.examples.simplecell

import pytest
import numpy

import deap.tools


@pytest.mark.unit
def test_DEAPOptimisation_constructor():
    "deapext.optimisation: Testing constructor DEAPOptimisation"

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    optimisation = bluepyopt.deapext.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator, map_function=map)

    assert isinstance(optimisation,
                      bluepyopt.deapext.optimisations.DEAPOptimisation)
    assert isinstance(optimisation.evaluator,
                      bluepyopt.evaluators.Evaluator)

    pytest.raises(
        ValueError,
        bluepyopt.deapext.optimisations.DEAPOptimisation,
        simplecell.cell_evaluator,
        selector_name='wrong')


@pytest.mark.unit
def test_IBEADEAPOptimisation_constructor():
    "deapext.optimisation: Testing constructor IBEADEAPOptimisation"

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    optimisation = bluepyopt.deapext.optimisations.IBEADEAPOptimisation(
        simplecell.cell_evaluator, map_function=map)

    assert isinstance(optimisation,
                      bluepyopt.deapext.optimisations.IBEADEAPOptimisation)


@pytest.mark.unit
def test_DEAPOptimisation_run():
    "deapext.optimisation: Testing DEAPOptimisation run"

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    optimisation = bluepyopt.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator, offspring_size=1)

    pop, hof, log, hist = optimisation.run(max_ngen=1)

    ind = [0.06007731830843009, 0.06508319290092013]
    assert len(pop) == 1
    numpy.testing.assert_almost_equal(pop[0], ind)
    numpy.testing.assert_almost_equal(hof[0], ind)
    assert log[0]['nevals'] == 1
    numpy.testing.assert_almost_equal(hist.genealogy_history[1], ind)


@pytest.mark.unit
def test_DEAPOptimisation_run_from_parents():
    "deapext.optimisation: Testing DEAPOptimisation run using prior parents"

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    optimisation = bluepyopt.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator, offspring_size=1)

    parent_population = [[0.060, 0.065]]
    pop, hof, log, hist = optimisation.run(max_ngen=0,
                                           parent_population=parent_population)

    assert len(pop) == 1
    numpy.testing.assert_almost_equal(pop[0], parent_population[0])


@pytest.mark.unit
def test_selectorname():
    "deapext.optimisation: Testing selector_name argument"

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()

    # Test default value
    ibea_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator)
    assert ibea_optimisation.selector_name == 'IBEA'

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()

    # Test NSGA2 selector
    nsga2_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator, selector_name='NSGA2')

    assert (
        nsga2_optimisation.toolbox.select.func
        == deap.tools.emo.selNSGA2)

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()

    # Test IBEA selector
    ibea_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        simplecell.cell_evaluator, selector_name='IBEA')

    assert (
        ibea_optimisation.toolbox.select.func
        == bluepyopt.deapext.tools.selIBEA)



"""bluepyopt.deapext.hype tests"""

import numpy

import bluepyopt.deapext.hype

import pytest


@pytest.mark.unit
def test_hypeIndicatorExact():
    """deapext.hype: Testing hypeIndicatorExact"""

    points = numpy.asarray([[250., 250.], [0., 0.], [240., 240.]])
    bounds = numpy.asarray([250., 250.])

    hv = bluepyopt.deapext.hype.hypeIndicatorExact(points, bounds, k=5)

    assert hv[0] == 0
    assert hv[1] == 62500
    assert hv[2] == 100


@pytest.mark.unit
def test_hypeIndicatorSampled():
    """deapext.hype: Testing hypeIndicatorSampled"""

    points = numpy.asarray([[250., 250.], [0., 0.], [240., 240.]])
    bounds = numpy.asarray([250., 250.])

    numpy.random.seed(42)
    hv = bluepyopt.deapext.hype.hypeIndicatorSampled(
        points, bounds, nrOfSamples=1000000, k=5
    )

    assert hv[0] == 0
    assert hv[1] == 62500
    assert numpy.abs((hv[2] / 100) - 1) < 0.05



"""Test bluepyopt.tools"""

import pytest


@pytest.mark.unit
def test_load():
    """bluepyopt.tools: test import"""

    import bluepyopt.tools  # NOQA


@pytest.mark.unit
def test_uint32_seed():
    """bluepyopt.tools: test uint32_seed"""

    import bluepyopt.tools as bpoptools

    assert bpoptools.uint32_seed("test") == 640136438

    import random
    random.seed(1)

    hashes = []
    strings = []
    for _ in range(1000):
        string = ''.join(
            (chr(random.randint(0, 127)) for x in
             range(random.randint(10, 255))))
        strings.append(string)
        hashes.append(bpoptools.uint32_seed(string))

    assert len(strings) == len(set(strings))
    assert len(hashes) == len(set(hashes))

    import numpy
    for hash_value in hashes:
        assert hash_value == numpy.uint32(hash_value)



"""bluepyopt.ephys.simulators tests"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612, W0201
import json
import numpy as np

import pytest

from bluepyopt import ephys
from bluepyopt.ephys.serializer import instantiator


@pytest.mark.unit
def test_location_init():
    """ephys.locations: test if Location works"""

    loc = ephys.locations.Location('test')
    assert isinstance(loc, ephys.locations.Location)
    assert loc.name == 'test'


@pytest.mark.unit
class TestNrnSectionCompLocation(object):

    """Test class for NrnSectionCompLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnSectionCompLocation(
            name='test',
            sec_name='soma[0]',
            comp_x=0.5)
        self.loc_dend = ephys.locations.NrnSectionCompLocation(
            name='test',
            sec_name='dend[1]',
            comp_x=0.5)
        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSomaDistanceCompLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):

            """Cell class"""
            pass
        cell = Cell()
        soma = self.sim.neuron.h.Section()
        dend1 = self.sim.neuron.h.Section(name='dend1')
        dend2 = self.sim.neuron.h.Section(name='dend2')

        cell.soma = [soma]
        cell.dend = [dend1, dend2]

        soma_comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert soma_comp == soma(0.5)

        dend_comp = self.loc_dend.instantiate(sim=self.sim, icell=cell)
        assert dend_comp == dend2(0.5)


@pytest.mark.unit
class TestNrnSeclistCompLocation(object):

    """Test class for NrnSectionCompLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnSeclistCompLocation(
            name='test',
            seclist_name='somatic',
            sec_index=0,
            comp_x=0.5)
        self.loc_dend = ephys.locations.NrnSeclistCompLocation(
            name='test',
            seclist_name='basal',
            sec_index=1,
            comp_x=0.5)
        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSeclistCompLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):
            """Cell class"""
            pass

        cell = Cell()
        soma = self.sim.neuron.h.Section()
        dend1 = self.sim.neuron.h.Section(name='dend1')
        dend2 = self.sim.neuron.h.Section(name='dend2')

        cell.somatic = self.sim.neuron.h.SectionList()
        cell.somatic.append(soma)
        cell.basal = self.sim.neuron.h.SectionList()
        cell.basal.append(dend1)
        cell.basal.append(dend2)

        soma_comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert soma_comp == soma(0.5)

        dend_comp = self.loc_dend.instantiate(sim=self.sim, icell=cell)
        assert dend_comp == dend2(0.5)

        for _ in range(10000):
            soma_comp = self.loc.instantiate(sim=self.sim, icell=cell)


@pytest.mark.unit
class TestNrnSeclistSecLocation(object):

    """Test class for NrnSeclistSecLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnSeclistSecLocation(
            name='test',
            seclist_name='somatic',
            sec_index=0)
        self.loc_dend = ephys.locations.NrnSeclistSecLocation(
            name='test',
            seclist_name='basal',
            sec_index=1)
        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSeclistSecLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):
            """Cell class"""
            pass

        cell = Cell()
        soma = self.sim.neuron.h.Section()
        dend1 = self.sim.neuron.h.Section(name='dend1')
        dend2 = self.sim.neuron.h.Section(name='dend2')

        cell.somatic = self.sim.neuron.h.SectionList()
        cell.somatic.append(soma)
        cell.basal = self.sim.neuron.h.SectionList()
        cell.basal.append(dend1)
        cell.basal.append(dend2)

        soma_comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert soma_comp == soma

        dend_comp = self.loc_dend.instantiate(sim=self.sim, icell=cell)
        assert dend_comp == dend2


@pytest.mark.unit
class TestNrnSomaDistanceCompLocation(object):

    """Test class for NrnSomaDistanceCompLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnSomaDistanceCompLocation(
            'test',
            125,
            'testdend')
        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSomaDistanceCompLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):

            """Cell class"""
            pass
        cell = Cell()
        soma = self.sim.neuron.h.Section()
        cell.soma = [soma]
        cell.testdend = self.sim.neuron.h.SectionList()
        dend1 = self.sim.neuron.h.Section(name='dend1')
        dend2 = self.sim.neuron.h.Section(name='dend2')

        cell.testdend.append(sec=dend1)
        cell.testdend.append(sec=dend2)

        pytest.raises(ephys.locations.EPhysLocInstantiateException,
                      self.loc.instantiate,
                      sim=self.sim,
                      icell=cell)

        dend1.connect(soma(0.5), 0.0)
        dend2.connect(dend1(1.0), 0.0)

        comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert comp == dend2(0.5)


@pytest.mark.unit
class TestNrnSecSomaDistanceCompLocation(object):

    """Test class for NrnSecSomaDistanceCompLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnSecSomaDistanceCompLocation(
            'test',
            125,
            1,
            'testdend')
        self.loc_other = ephys.locations.NrnSecSomaDistanceCompLocation(
            'test',
            250,
            4,
            'testdend')
        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSomaDistanceCompLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):

            """Cell class"""
            pass
        cell = Cell()
        soma = self.sim.neuron.h.Section(name="soma[0]")
        cell.soma = [soma]
        cell.testdend = self.sim.neuron.h.SectionList()
        dend1 = self.sim.neuron.h.Section(name='dend[0]')
        dend2 = self.sim.neuron.h.Section(name='dend[1]')
        dend3 = self.sim.neuron.h.Section(name='dend[2]')
        dend4 = self.sim.neuron.h.Section(name='dend[3]')
        dend5 = self.sim.neuron.h.Section(name='dend[4]')

        cell.testdend.append(sec=dend1)
        cell.testdend.append(sec=dend2)
        cell.testdend.append(sec=dend3)
        cell.testdend.append(sec=dend4)
        cell.testdend.append(sec=dend5)

        dend1.connect(soma(0.5), 0.0)
        dend2.connect(dend1(1.0), 0.0)
        dend3.connect(dend1(1.0), 0.0)
        dend4.connect(dend3(1.0), 0.0)
        dend5.connect(dend4(1.0), 0.0)

        comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert comp == dend2(0.5)
        comp = self.loc_other.instantiate(sim=self.sim, icell=cell)
        assert comp == dend4(0.5)


@pytest.mark.unit
class TestNrnTrunkSomaDistanceCompLocation(object):

    """Test class for NrnTrunkSomaDistanceCompLocation"""

    def setup(self):
        """Setup"""
        self.loc = ephys.locations.NrnTrunkSomaDistanceCompLocation(
            'test',
            soma_distance=150,
            seclist_name='testdend')
        self.loc_other = ephys.locations.NrnTrunkSomaDistanceCompLocation(
            'test',
            soma_distance=350,
            seclist_name='testdend')

        assert self.loc.name == 'test'
        self.sim = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """ephys.locations.NrnSomaDistanceCompLocation: test instantiate"""

        # Create a little test class with a soma and two dendritic sections
        class Cell(object):

            """Cell class"""
            pass
        cell = Cell()
        soma = self.sim.neuron.h.Section(name="soma[0]")
        cell.soma = [soma]
        cell.testdend = self.sim.neuron.h.SectionList()
        dend1 = self.sim.neuron.h.Section(name='dend[0]')
        dend2 = self.sim.neuron.h.Section(name='dend[1]')
        dend3 = self.sim.neuron.h.Section(name='dend[2]')
        dend4 = self.sim.neuron.h.Section(name='dend[3]')
        dend5 = self.sim.neuron.h.Section(name='dend[4]')

        cell.testdend.append(sec=dend1)
        cell.testdend.append(sec=dend2)
        cell.testdend.append(sec=dend3)
        cell.testdend.append(sec=dend4)
        cell.testdend.append(sec=dend5)

        x0 = self.sim.neuron.h.Vector([0] * 10)
        d = self.sim.neuron.h.Vector([1] * 10)

        x1 = self.sim.neuron.h.Vector(np.linspace(0, 100, 10))
        self.sim.neuron.h.pt3dadd(x0, x1, x0, d, sec=dend1)
        x2 = self.sim.neuron.h.Vector(np.linspace(100, 200, 10))
        self.sim.neuron.h.pt3dadd(x0, x2, x0, d, sec=dend2)
        x3 = self.sim.neuron.h.Vector(np.linspace(200, 300, 10))
        self.sim.neuron.h.pt3dadd(x0, x3, x0, d, sec=dend3)
        x4 = self.sim.neuron.h.Vector(np.linspace(300, 400, 10))
        self.sim.neuron.h.pt3dadd(x0, x4, x0, d, sec=dend4)
        x5 = self.sim.neuron.h.Vector(np.linspace(400, 500, 10))
        self.sim.neuron.h.pt3dadd(x0, x5, x0, d, sec=dend5)

        dend1.connect(soma(0.5), 0.0)
        dend2.connect(dend1(1.0), 0.0)
        dend3.connect(dend1(1.0), 0.0)
        dend4.connect(dend3(1.0), 0.0)
        dend5.connect(dend4(1.0), 0.0)

        comp = self.loc.instantiate(sim=self.sim, icell=cell)
        assert comp == dend3(0.5)
        comp = self.loc_other.instantiate(sim=self.sim, icell=cell)
        assert comp == dend5(0.5)


@pytest.mark.unit
def test_serialize():
    """ephys.locations: Test serialize functionality"""
    from bluepyopt.ephys.locations import (
        NrnSeclistCompLocation,
        NrnSeclistLocation,
        NrnSeclistSecLocation,
        NrnSomaDistanceCompLocation)

    seclist_name, sec_index, comp_x, soma_distance = 'somatic', 0, 0.5, 800
    locations = (
        NrnSeclistCompLocation('NrnSeclistCompLocation',
                               seclist_name, sec_index, comp_x),
        NrnSeclistLocation(
            'NrnSeclistLocation', seclist_name),
        NrnSeclistSecLocation(
            'NrnSeclistSecLocation', seclist_name, sec_index),
        NrnSomaDistanceCompLocation(
            'NrnSomaDistanceCompLocation', soma_distance, seclist_name),)

    for loc in locations:
        serialized = loc.to_dict()
        assert isinstance(json.dumps(serialized), str)
        deserialized = instantiator(serialized)
        assert isinstance(deserialized, loc.__class__)
        assert deserialized.seclist_name == seclist_name
        assert deserialized.name == loc.__class__.__name__



"""Create l5pc model with MEA"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import sys
import os
import MEAutility as mu

import bluepyopt
import bluepyopt.ephys as ephys

L5PC_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '../l5pc'))
sys.path.insert(0, L5PC_PATH)

import l5pc_model


def define_electrode(
    probe_center=[0, 300, 20],
    mea_dim=[10, 4],
    mea_pitch=[300, 300],
):
    """
    Defines LFPy electrode object
    Parameters
    ----------
    probe_center: 3d array
        The center of the probe
    mea_dim: 2d array
        Dimensions of planar probe (nrows, ncols)
    mea_pitch: 3d arraay
        The pitch of the planar probe (row pitch, column pitch)
    Returns
    -------
    electrode: MEAutility.MEA object
        The MEAutility electrode object
    """

    mea_info = {
        'dim': mea_dim,
        'electrode_name': 'hd-mea',
        'pitch': mea_pitch,
        'shape': 'square',
        'size': 5,
        'type': 'mea',
        'plane': 'xy'
    }
    probe = mu.return_mea(info=mea_info)

    # Move the MEA out of the neuron plane (yz)
    probe.move(probe_center)

    return probe


def create():

    l5pc_cell = ephys.models.LFPyCellModel(
        "l5pc_lfpy",
        morph=l5pc_model.define_morphology(do_replace_axon=True),
        mechs=l5pc_model.define_mechanisms(),
        params=l5pc_model.define_parameters(),
        electrode=define_electrode()
    )

    return l5pc_cell



"""bluepyopt.optimisationsCMA tests"""

import pytest
import bluepyopt
import bluepyopt.ephys.examples.simplecell


@pytest.mark.unit
def test_optimisationsCMA_normspace():
    """deapext.optimisationsCMA: Testing optimisationsCMA normspace"""

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    evaluator = simplecell.cell_evaluator
    optimisation = bluepyopt.deapext.optimisationsCMA.DEAPOptimisationCMA(
        evaluator=evaluator)

    x = [n * 0.1 for n in range(len(evaluator.params))]
    y = [f2(f1(p)) for p, f1, f2 in zip(x, optimisation.to_norm,
                                        optimisation.to_space)]

    for a, b in zip(x, y):
        assert b == pytest.approx(a, abs=1e-5)


@pytest.mark.unit
def test_optimisationsCMA_SO_run():
    """deapext.optimisationsCMA: Testing optimisationsCMA run from centroid"""

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    evaluator = simplecell.cell_evaluator
    x = [n * 0.1 for n in range(len(evaluator.params))]

    optimiser = bluepyopt.deapext.optimisationsCMA.DEAPOptimisationCMA
    optimisation = optimiser(evaluator=evaluator, centroids=[x])
    pop, hof, log, hist = optimisation.run(max_ngen=2)

    assert log.select("avg")[-1] == pytest.approx(53.3333, abs=1e-4)
    assert log.select("std")[-1] == pytest.approx(83.7987, abs=1e-4)
    assert pop[0][0] == pytest.approx(0.10525059698894745, abs=1e-6)
    assert pop[0][1] == pytest.approx(0.01000000003249999, abs=1e-6)


@pytest.mark.unit
def test_optimisationsCMA_MO_run():
    """deapext.optimisationsCMA: Testing optimisationsCMA run from centroid"""

    simplecell = bluepyopt.ephys.examples.simplecell.SimpleCell()
    evaluator = simplecell.cell_evaluator

    optimiser = bluepyopt.deapext.optimisationsCMA.DEAPOptimisationCMA
    optimisation = optimiser(
        selector_name="multi_objective",
        offspring_size=3,
        evaluator=evaluator,
        seed=42
    )
    pop, hof, log, hist = optimisation.run(max_ngen=2)

    assert log.select("avg")[-1] == pytest.approx(40., abs=1e-4)
    assert log.select("std")[-1] == pytest.approx(16.32993, abs=1e-4)
    assert pop[0][0] == pytest.approx(0.09601241274168831, abs=1e-6)
    assert pop[0][1] == pytest.approx(0.024646650865379722, abs=1e-6)



"""Location classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import itertools

from bluepyopt.ephys.base import BaseEPhys
from bluepyopt.ephys.serializer import DictMixin
from bluepyopt.ephys.parameterscalers import format_float
from bluepyopt.ephys.acc import ArbLabel
from bluepyopt.ephys.morphologies import ArbFileMorphology
import numpy as np

import logging
logger = logging.getLogger(__name__)


class Location(BaseEPhys):

    """Location"""
    pass

# TODO make all these locations accept a cell name
# TODO instantiate should get the entire simulation environment
# TODO find better/more general name for this
# TODO specify in document abrevation comp=compartment, sec=section, ...


def _nth_isectionlist(isectionlist, index):
    """Get nth element of isectionlist

    Sectionlists don't support direct indexing
    """
    isection = next(
        itertools.islice(
            isectionlist,
            index,
            index + 1))
    return isection


class NrnSeclistCompLocation(Location, DictMixin):

    """Compartment in a sectionlist"""

    SERIALIZED_FIELDS = (
        'name',
        'comment',
        'seclist_name',
        'sec_index',
        'comp_x',
    )

    def __init__(
            self,
            name,
            seclist_name=None,
            sec_index=None,
            comp_x=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of the object
            seclist_name (str): name of Neuron section list (ex: 'somatic')
            sec_index (int): index of the section in the section list
            comp_x (float): segx (0..1) of segment inside section
        """

        super(NrnSeclistCompLocation, self).__init__(name, comment)
        self.seclist_name = seclist_name
        self.sec_index = sec_index
        self.comp_x = comp_x

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiate compartment"""
        iseclist = getattr(icell, self.seclist_name)

        iseclist_size = len([x for x in iseclist])
        if self.sec_index >= iseclist_size:
            raise Exception(
                'NrnSeclistCompLocation: section index %d falls out of '
                'SectionList size of %d' %
                (self.sec_index, iseclist_size))
        isection = _nth_isectionlist(iseclist, self.sec_index)
        icomp = isection(self.comp_x)

        # The code above seems to add put a section on the stack
        # TODO remove line below once we figure out where the section is pushed
        sim.neuron.h.pop_section()

        return icomp

    def acc_label(self):
        """Arbor label"""
        raise EPhysLocAccException(
            '%s not supported in Arbor' % type(self).__name__ +
            ' (uses branches instead of NEURON sections).'
            ' Use ArbBranchRelLocation/ArbSegmentRelLocation/'
            'ArbLocsetLocation instead (consider using the'
            ' Arbor GUI to identify the precise branch/segment index'
            ' and relative position).')

    def __str__(self):
        """String representation"""

        return '%s[%s](%s)' % (self.seclist_name, self.sec_index, self.comp_x)


class NrnSectionCompLocation(Location, DictMixin):

    """Compartment in a section"""

    SERIALIZED_FIELDS = (
        'name',
        'comment',
        'seclist_name',
        'sec_index',
        'comp_x',
    )

    def __init__(
            self,
            name,
            sec_name=None,
            comp_x=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of the object
            sec_name (str): name of Neuron section (ex: 'soma[0]')
            comp_x (float): segx (0..1) of segment inside section
        """

        super(NrnSectionCompLocation, self).__init__(name, comment)
        self.sec_name = sec_name
        self.comp_x = comp_x

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiate compartment"""

        # Dont see any other way but to use eval, apart from parsing the
        # sec_name string which can be complicated
        isection = eval('icell.%s' % self.sec_name)  # pylint: disable=W0123

        icomp = isection(self.comp_x)
        return icomp

    def acc_label(self):
        """Arbor label"""
        raise EPhysLocAccException(
            '%s not supported in Arbor' % type(self).__name__ +
            ' (uses branches instead of NEURON sections).'
            ' Use ArbBranchRelLocation/ArbSegmentRelLocation/'
            'ArbLocsetLocation instead (consider using the'
            ' Arbor GUI to identify the precise branch/segment index'
            ' and relative position).')

    def __str__(self):
        return '%s(%s)' % (self.sec_name, self.comp_x)


class NrnPointProcessLocation(Location):

    """Point process location"""

    def __init__(
            self,
            name,
            pprocess_mech,
            comment=''):
        """Constructor

        Args:
            name (str): name of the object
            pprocess_mech (str): point process mechanism
        """

        super(NrnPointProcessLocation, self).__init__(name, comment)
        self.pprocess_mech = pprocess_mech

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiated point processes"""

        return self.pprocess_mech.pprocesses

    def acc_label(self):
        """Arbor label"""
        return [loc.acc_label() for loc in self.pprocess_mech.locations]

    def __str__(self):
        """String representation"""

        return '%s' % (self.pprocess_mech.name)


class NrnSeclistLocation(Location, DictMixin):

    """Section in a sectionlist"""

    SERIALIZED_FIELDS = ('name', 'comment', 'seclist_name', )

    def __init__(
            self,
            name,
            seclist_name=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of the object
            seclist_name (str): name of NEURON section list (ex: 'somatic')
        """

        super(NrnSeclistLocation, self).__init__(name, comment)
        self.seclist_name = seclist_name

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiate compartment"""

        isectionlist = getattr(icell, self.seclist_name)

        return (isection for isection in isectionlist)

    def acc_label(self):
        """Arbor label"""
        return ArbFileMorphology.region_labels[self.seclist_name]

    def __str__(self):
        """String representation"""

        return '%s' % (self.seclist_name)


class NrnSeclistSecLocation(Location, DictMixin):

    """Section in a sectionlist"""

    SERIALIZED_FIELDS = ('name', 'comment', 'seclist_name', 'sec_index', )

    def __init__(
            self,
            name,
            seclist_name=None,
            sec_index=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of this object
            seclist_name (str): name of Neuron section list (ex: 'somatic')
            sec_index (int): index of the section
        """

        super(NrnSeclistSecLocation, self).__init__(name, comment)
        self.seclist_name = seclist_name
        self.sec_index = sec_index

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiate compartment"""

        isectionlist = getattr(icell, self.seclist_name)
        isection = _nth_isectionlist(isectionlist, self.sec_index)
        return isection

    def acc_label(self):
        """Arbor label"""
        raise EPhysLocAccException(
            '%s not supported in Arbor' % type(self).__name__ +
            ' (uses branches instead of NEURON sections).'
            ' Use ArbBranchLocation/ArbSegmentLocation/ArbRegionLocation'
            ' instead (consider using the Arbor GUI to identify the'
            ' precise branch/segment index).')

    def __str__(self):
        """String representation"""

        return '%s[%s]' % (self.seclist_name, self.sec_index)


class NrnSomaDistanceCompLocation(Location, DictMixin):

    """Compartment at distance from soma"""

    SERIALIZED_FIELDS = ('name', 'comment', 'soma_distance', 'seclist_name', )

    def __init__(
            self,
            name,
            soma_distance=None,
            seclist_name=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of this object
            soma_distance (float): distance from soma to this compartment
            seclist_name (str): name of Neuron section list (ex: 'apical')
        """

        super(NrnSomaDistanceCompLocation, self).__init__(name, comment)
        self.soma_distance = soma_distance
        self.seclist_name = seclist_name

    # TODO this definitely has to be unit-tested
    # TODO add ability to specify origin
    def find_icomp(self, sim, iseclist):
        """Find the index of the compartment based on a list of isec
           and a distance"""
        icomp = None

        for isec in iseclist:
            start_distance = sim.neuron.h.distance(1, 0.0, sec=isec)
            end_distance = sim.neuron.h.distance(1, 1.0, sec=isec)

            min_distance = min(start_distance, end_distance)
            max_distance = max(start_distance, end_distance)

            if min_distance <= self.soma_distance <= max_distance:
                comp_x = float(self.soma_distance - min_distance) / \
                    (max_distance - min_distance)

                comp_diam = isec(comp_x).diam

                if comp_diam > 0.0:
                    icomp = isec(comp_x)

        if icomp is None:
            raise EPhysLocInstantiateException(
                'No comp found at %s distance from soma' %
                self.soma_distance)

        return icomp

    def instantiate(self, sim=None, icell=None):
        """Find the instantiate compartment"""

        soma = icell.soma[0]

        sim.neuron.h.distance(0, 0.5, sec=soma)

        iseclist = getattr(icell, self.seclist_name)

        return self.find_icomp(sim, iseclist)

    def acc_label(self):
        """Arbor label"""
        # Potentially non-unique location - in that case to be refined in the
        # Arbor GUI (create ArbLocsetLocation directly).
        # Alternatives to (on-components 0.5 (region "soma")) are
        #  - '(segment <id>)'
        #  - '(proximal (region %s)))' % self.seclist_name
        # If outer restrict results in non-unique location (cf. GUI) use
        # specific branch or similar instead of seclist_name, e.g.
        #  - (proximal-interval (distal (branch <id>)))
        # for a branch distally from the desired location
        acc_label = ArbLabel(
            'locset', self.name,
            '(restrict (distal-translate (on-components 0.5 %s) %s) %s)' %
            (ArbFileMorphology.region_labels['somatic'].ref,
             format_float(self.soma_distance),
             ArbFileMorphology.region_labels[self.seclist_name].ref))
        logger.warning(
            'Make sure that ACC label %s' % acc_label.loc +
            ' for NrnSomaDistanceCompLocation (%s) ' % str(self) +
            ' instantiates to a unique location on the morphology.'
            ' Use the Arbor GUI to validate/refine the location expression.')
        return acc_label

    def __str__(self):
        """String representation"""

        return '%f micron from soma in %s' % (
            self.soma_distance, self.seclist_name)


class NrnSecSomaDistanceCompLocation(NrnSomaDistanceCompLocation):

    """Compartment on a section defined both by a section index and distance
       from the soma """

    SERIALIZED_FIELDS = ('name', 'comment', 'soma_distance', 'seclist_name',
                         'sec_index', )

    def __init__(
        self,
        name,
        soma_distance=None,
        sec_index=None,
        seclist_name=None,
        comment=""
    ):
        """Constructor

        Args:
            name (str): name of this object
            soma_distance (float): distance from soma to this compartment
            sec_index (int): index of the section  to consider
            seclist_name (str): name of Neuron sections (ex: 'apic')
        """
        super(NrnSecSomaDistanceCompLocation, self).__init__(
            name,
            soma_distance=soma_distance,
            seclist_name=seclist_name,
            comment=comment,
        )
        self.sec_index = sec_index

    def instantiate(self, sim=None, icell=None):
        """Find the instantiate compartment"""

        if self.sec_index is None:
            raise EPhysLocInstantiateException(
                "No apical point was given")

        sections = getattr(icell, self.seclist_name)
        section = _nth_isectionlist(sections, self.sec_index)

        branches = []
        while True:

            name = str(section.name()).split(".")[-1]
            if name == "soma[0]":
                break

            branches.append(section)

            if sim.neuron.h.SectionRef(sec=section).has_parent():
                section = sim.neuron.h.SectionRef(sec=section).parent
            else:
                raise EPhysLocInstantiateException(
                    "soma[0] was not reached from isec point "
                    "%f" % self.sec_index
                )

        soma = icell.soma[0]

        sim.neuron.h.distance(0, 0.5, sec=soma)

        return self.find_icomp(sim, branches)

    def acc_label(self):
        """Arbor label"""
        raise EPhysLocAccException('%s not supported in Arbor.' %
                                   type(self).__name__)


class NrnTrunkSomaDistanceCompLocation(NrnSecSomaDistanceCompLocation):
    """Location at a distance from soma along a main direction.

    We search for the section that is the furthest away from some along
    a direction, and pick a location at a given distance from soma along
    the path to that section.

    If direction == 'radial', the largest radial direction is used.

    This is most useful to follow the trunk of an apical dendrite
    without knowing the apical point, but only that apical trunk goes along y.
    """

    def __init__(
        self,
        name,
        soma_distance=None,
        sec_index=None,
        seclist_name=None,
        direction=None,
        comment=""
    ):
        """Constructor

        Args:
            name (str): name of this object
            soma_distance (float): distance from soma to this compartment
            sec_index (int): index of the section  to consider
            sec_name (str): name of Neuron sections (ex: 'apic')
            direction (list of 3 elements): 3d vector representing direction,
                if None, default is [0, 1, 0]
        """
        super(NrnTrunkSomaDistanceCompLocation, self).__init__(
            name,
            soma_distance=soma_distance,
            sec_index=sec_index,
            seclist_name=seclist_name,
            comment=comment
        )

        if direction is None:
            direction = [0.0, 1.0, 0.0]
        self.direction = direction

    def set_sec_index(self, icell=None):
        """Search for the point furthest away along given direction."""
        points = np.array(
            [
                [
                    section.x3d(section.n3d() - 1),
                    section.y3d(section.n3d() - 1),
                    section.z3d(section.n3d() - 1),
                ]
                for section in getattr(icell, self.seclist_name)
            ]
        )
        if len(points) > 0:
            if self.direction == 'radial':
                self.sec_index = int(np.argmax(np.linalg.norm(points, axis=1)))
            else:
                self.sec_index = int(np.argmax(points.dot(self.direction)))
        else:
            raise EPhysLocInstantiateException(
                "Empty seclist: %s" % self.seclist_name
            )

    def instantiate(self, sim=None, icell=None):
        """ """
        if self.sec_index is None:
            self.set_sec_index(icell=icell)
        return super().instantiate(sim=sim, icell=icell)

    def acc_label(self):
        """Arbor label"""
        raise EPhysLocAccException('%s not supported in Arbor.' %
                                   type(self).__name__)


class ArbLocation(Location):
    """Arbor Location"""

    def instantiate(self, sim=None, icell=None):  # pylint: disable=W0613
        """Find the instantiate compartment (default implementation)"""
        raise EPhysLocInstantiateException(
            '%s not supported in NEURON.' % type(self).__name__)

    def __str__(self):
        """String representation"""
        return '%s \'%s\'' % (type(self).__name__, self.acc_label().defn)


class ArbSegmentLocation(ArbLocation):
    """Segment in an Arbor morphology.
    """

    def __init__(self, name, segment, comment=''):
        super().__init__(name, comment)
        self.segment = segment

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('region', self.name, '(segment %s)' % (self.segment))


class ArbBranchLocation(ArbLocation):
    """Branch in an Arbor morphology.

    Arbor's counterpart of NrnSeclistSecLocation.
    """

    def __init__(self, name, branch, comment=''):
        super().__init__(name, comment)
        self.branch = branch

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('region', self.name, '(branch %s)' % (self.branch))


class ArbSegmentRelLocation(ArbLocation):
    """Relative position on a segment in an Arbor morphology.
    """

    def __init__(self, name, segment, pos, comment=''):
        super().__init__(name, comment)
        self.segment = segment
        self.pos = pos

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('locset', self.name,
                        '(on-components %s (segment %s))' %
                        (format_float(self.pos), self.segment))


class ArbBranchRelLocation(ArbLocation):
    """Relative position on a branch in an Arbor morphology.

    Arbor's counterpart of NrnSeclistCompLocation.
    """

    def __init__(self, name, branch, pos, comment=''):
        super().__init__(name, comment)
        self.branch = branch
        self.pos = pos

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('locset', self.name,
                        '(location %s %s)' %
                        (self.branch, format_float(self.pos)))


class ArbLocsetLocation(ArbLocation):
    """Arbor location set defined by a user-supplied string (S-expression).
    """

    def __init__(self, name, locset, comment=''):
        super().__init__(name, comment)
        self.locset = locset

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('locset', self.name, self.locset)


class ArbRegionLocation(ArbLocation):
    """Arbor region defined by a user-supplied string (S-expression).
    """

    def __init__(self, name, region, comment=''):
        super().__init__(name, comment)
        self.region = region

    def acc_label(self):
        """Arbor label"""
        return ArbLabel('region', self.name, self.region)


class EPhysLocInstantiateException(Exception):

    """All exceptions generated by location instantiation"""

    def __init__(self, message):
        """Constructor"""

        super(EPhysLocInstantiateException, self).__init__(message)


class EPhysLocAccException(Exception):

    """All exceptions generated by ACC label creation"""

    def __init__(self, message):
        """Constructor"""

        super(EPhysLocAccException, self).__init__(message)



'''Dependencies of Arbor simulator backend'''

try:
    import arbor
except ImportError as e:
    class arbor:
        def __getattribute__(self, _):
            raise ImportError("Exporting cell models to ACC/JSON, loading"
                              " them or optimizing them with the Arbor"
                              " simulator requires missing dependency arbor."
                              " To install BluePyOpt with arbor,"
                              " run 'pip install bluepyopt[arbor]'.")


class ArbLabel:
    """Arbor label"""

    def __init__(self, type, name, s_expr):
        if type not in ['locset', 'region', 'iexpr']:
            raise ValueError('Invalid Arbor label type %s' % type)
        self._type = type
        self._name = name
        self._s_expr = s_expr

    @property
    def defn(self):
        """Label definition for label-dict"""
        return '(%s-def "%s" %s)' % (self._type, self._name, self._s_expr)

    @property
    def ref(self):
        """Reference to label defined in label-dict"""
        return '(%s "%s")' % (self._type, self._name)

    @property
    def name(self):
        """Name of the label"""
        return self._name

    @property
    def loc(self):
        """S-expression defining the location of the label"""
        return self._s_expr

    def __eq__(self, other):
        if other is None:
            return False
        elif not isinstance(other, ArbLabel):
            raise TypeError('%s is not an ArbLabel' % str(other))
        else:
            return self._s_expr == other._s_expr

    def __hash__(self):
        return hash(self._s_expr)

    def __repr__(self):
        return self.defn



"""Functions to create neuroml biophysiology"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project
 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>
 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.
 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.
 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil
from pathlib import Path

import neuroml

from bluepyopt import ephys

ignore_params = ["cm", "Ra", "ena", "ek"]

# Due to the use of 1e-4 in BREAKPOINT in StochKv.mod:
# ik = 1e-4 * gk * (v - ek)
density_scales = {"StochKv": 1e-4}

channel_substitutes = {"StochKv": "StochKv_deterministic"}

channel_ions = {
    "Ih": "hcn",
    "NaTa_t": "na",
    "NaTs2_t": "na",
    "Nap_Et2": "na",
    "K_Tst": "k",
    "K_Pst": "k",
    "SKv3_1": "k",
    "SK_E2": "k",
    "StochKv": "k",
    "KdShu2007": "k",
    "Im": "k",
    "Ca": "ca",
    "Ca_HVA": "ca",
    "Ca_LVAst": "ca",
    "pas": "pas",
    "CaDynamics_E2": "ca",
}

ion_erevs = {
    "na": "50.0 mV",
    "k": "-85.0 mV",
    "hcn": "-45.0 mV",
    "ca": "nernst",
    "pas": "pas",
}

default_capacitances = {
    "axonal": "1.0 uF_per_cm2",
    "somatic": "1.0 uF_per_cm2",
    "basal": "1.0 uF_per_cm2",
    "apical": "1.0 uF_per_cm2",
}


def get_nml_mech_dir():
    """Returns path to repo containing neuroml mechanisms."""
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), "NeuroML2_mechanisms")
    )


def adapt_CaDynamics_nml(
    new_concentrations,
    channel_dir="channels",
):
    """Write concentration model if not present in CaDynamics nml file.

    We have to use this non-neuroml2 approved trick until
    https://github.com/NeuroML/NeuroML2/issues/153 is solved.

    Arguments:
        new_concentrations (dict): dict of the form
            "model_name": {"gamma": value "decay": value}
            describing the concentrations to append to
            ./{channel_dir}/CaDynamics_E2_NML2.nml
        channel_dir (str): repo in which to copy the channel files
    """
    place_after = (
        '    <concentrationModel id="CaDynamics_E2_NML2" '
        'type="concentrationModelHayEtAl" '
        'minCai="1e-4 mM" decay="80 ms" depth="0.1 um" '
        'gamma="0.05" ion="ca"/>\n'
    )
    cadyn_filename = "baseCaDynamics_E2_NML2.nml"
    new_cadyn_filename = "CaDynamics_E2_NML2.nml"
    nml_mech_dir = get_nml_mech_dir()

    with open(Path(nml_mech_dir) / cadyn_filename, "r") as f:
        lines = f.readlines()

    for model, value_dict in new_concentrations.items():
        new_concentration = (
            f'    <concentrationModel id="{model}" ion="ca" '
            'type="concentrationModelHayEtAl" minCai="1e-4 mM" '
            f'gamma="{value_dict["gamma"]}" '
            f'decay="{value_dict["decay"]} ms" depth="0.1 um"/>\n'
        )
        if new_concentration not in lines:
            idx = lines.index(place_after) + 2  # 2 to account for blank line
            lines.insert(idx, f"{new_concentration}\n")

    Path(channel_dir).mkdir(exist_ok=True)
    with open(Path(channel_dir) / new_cadyn_filename, "w") as f:
        contents = "".join(lines)
        f.write(contents)


def get_channel_from_param_name(param_name):
    """Return channel name, given parameter name.

    Arguments:
        param_name (str): parameter name used within NEURON
    """
    split_name = param_name.split("_")
    if len(split_name) == 4:
        channel = "_".join(split_name[2:4])
    elif len(split_name) == 3:
        channel = "_".join(split_name[1:3])
    elif len(split_name) == 2:
        channel = split_name[1]
    else:
        raise Exception(
            f"Could not extract channel from parameter name {param_name}"
        )

    return channel


def format_dist_fun(raw_expr, value, dist_param_names):
    """Format and return the distribution expression.

    Arguments:
        raw_expr (str): the function expression to be formatted
        value (float): the value to be put in the expression
        dist_param_names (list): list of names of parameters that parametrise
            the distribution
    """
    if dist_param_names is not None:
        raise NotImplementedError(
            "Functions that depend on other parameters, "
            "like decay function, are not implemented yet."
        )
    new_expr = raw_expr.format(distance="p", value=value)
    if "math" in new_expr:
        new_expr = new_expr.replace("math.", "")
    if "(p)" in new_expr:
        new_expr = new_expr.replace("(p)", "p")

    return new_expr


def add_nml_channel_to_nml_cell_file(
    cell_doc,
    included_channels,
    channel_name=None,
    channel_nml2_file=None,
    channel_dir="channels",
    skip_channels_copy=False,
):
    """Add NeuroML channel file to NeuroML cell file.

    And copy channel in the current directory.

    Arguments:
        cell_doc (NeuroMLDocument): nml document of the cell model
        included_channels (list): list of already included channels
        channel_name (str): name of the channel
        channel_nml2_file (str): name of the neuroml channel file
        channel_dir (str): repo in which to copy the channel files
        skip_channels_copy (bool): True to skip the copy pasting
            of the neuroml channel files
    """
    if channel_nml2_file is None:
        if channel_name is None:
            raise ValueError(
                "Plaise provide either a channel name or a channel nml2 file."
            )
        channel_nml2_file = f"{channel_name}.channel.nml"

    if channel_nml2_file not in included_channels:
        channel_new_path = Path(channel_dir) / channel_nml2_file
        cell_doc.includes.append(neuroml.IncludeType(href=channel_new_path))

        # for some reason, pynml cannot accept absolute paths in IncludeType,
        # so copy paste files in current directory for the simulation to work
        if not skip_channels_copy:
            Path(channel_dir).mkdir(exist_ok=True)
            nml_mech_dir = Path(get_nml_mech_dir())
            channel_path = nml_mech_dir / channel_nml2_file
            if channel_path.is_file():
                shutil.copy(channel_path, channel_new_path)

        included_channels.append(channel_nml2_file)


def get_channel_ion(channel, custom_channel_ion=None):
    """Get ion name given channel name.

    Arguments:
        channel (str): ion channel (e.g. StochKv)
        custom_channel_ion (dict): dict mapping channel to ion
    """
    ion = channel_ions.get(channel, None)
    if ion is None and custom_channel_ion is not None:
        ion = custom_channel_ion.get(channel, None)
    if ion is None:
        raise KeyError(
            f"Ion not found for channel {channel}."
            " Please set channel-ion mapping using custom_channel_ion."
        )
    return ion


def get_erev(ion, custom_ion_erevs=None):
    """Get reversal potential as str given ion name.

    Arguments:
        ion (str): ion name (e.g. na)
        custom_ion_erevs (dict): dict mapping ion to erev (reversal potential)
    """
    erev = ion_erevs.get(ion, None)
    if erev is None and custom_ion_erevs is not None:
        erev = custom_ion_erevs.get(ion, None)
    if erev is None:
        raise KeyError(
            f"Reversal potential not found for ion {ion}."
            " Please set ion-erev mapping using custom_ion_erevs."
        )
    return erev


def get_arguments(
    params,
    parameter_name,
    section_list,
    channel,
    channel_name,
    variable_parameters,
    cond_density,
    release_params,
    custom_channel_ion=None,
    custom_ion_erevs=None,
):
    """Get arguments for channel density function.

    Arguments:
        params (dict): contains the cell's parameters
        parameter_name (str): name of the parameter (e.g. e_pas)
        section_list (str): name of the location of the parameter (e.g. axonal)
        channel (str): ion channel (e.g. StochKv)
        channel_name (str): ion channel name used in the neuroML channel file
            (e.g. StochKv_deterministic)
        variable_parameters (list of neuroml.VariableParameter):
            parameters for non-uniform distributions
        cond_density (str): conductance density
        release_params (dict): optimized parameters
        custom_channel_ion (dict): dict mapping channel to ion
        custom_ion_erevs (dict): dict mapping ion to erev (reversal potential)
    """
    arguments = {}

    arguments["ion"] = get_channel_ion(channel, custom_channel_ion)
    erev = get_erev(arguments["ion"], custom_ion_erevs)

    channel_class = "ChannelDensity"

    if erev == "nernst":
        erev = None
        channel_class = "ChannelDensityNernst"
    elif erev == "pas":
        erev = params[f"e_pas.{section_list}"].value
        if erev is None:
            # non frozen parameter
            erev = release_params[f"e_pas.{section_list}"]
        erev = f"{erev} mV"
        arguments["ion"] = "non_specific"

    if variable_parameters is not None:
        channel_class += "NonUniform"
    else:
        arguments["segment_groups"] = section_list

    if erev is not None:
        arguments["erev"] = erev
    arguments["id"] = f"{section_list}_{parameter_name}"
    if cond_density is not None:
        arguments["cond_density"] = cond_density
    arguments["ion_channel"] = channel_name
    if variable_parameters is not None:
        arguments["variable_parameters"] = variable_parameters

    return arguments, channel_class


def extract_parameter_value(
    parameter, section_list, channel, skip_non_uniform, release_params
):
    """Returns conductance density and variable parameters.

    Arguments:
        parameter (ephys.parameters)
        section_list (str): location
        channel (str): ion channel
        skip_non_uniform (bool): True to skip non uniform distributions
        release_params (dict): optimized parameters
    """
    cond_density = None
    variable_parameters = None

    # uniform
    if isinstance(
        parameter.value_scaler, ephys.parameterscalers.NrnSegmentLinearScaler
    ):
        value = parameter.value
        if value is None:
            # non frozen parameter
            value = release_params[parameter.name]
        if channel in density_scales:
            value = value * density_scales[channel]
        cond_density = f"{value} S_per_cm2"
    # non uniform
    elif not skip_non_uniform:
        value = parameter.value
        if value is None:
            # non frozen parameter
            value = release_params[parameter.name]
        # did not mulyiply by 1e4. Is that ok?
        new_expr = format_dist_fun(
            raw_expr=parameter.value_scaler.distribution,
            value=value,
            dist_param_names=parameter.value_scaler.dist_param_names,
        )

        iv = neuroml.InhomogeneousValue(
            inhomogeneous_parameters=f"PathLengthOver_{section_list}",
            value=new_expr,
        )
        variable_parameters = [
            neuroml.VariableParameter(
                segment_groups=section_list,
                parameter="condDensity",
                inhomogeneous_value=iv,
            )
        ]
    else:
        return None, None

    return cond_density, variable_parameters


def get_density(
    cell_doc,
    cell,
    parameter,
    section_list,
    included_channels,
    skip_non_uniform,
    release_params,
    skip_channels_copy,
    custom_channel_ion=None,
    custom_ion_erevs=None,
):
    """Return density.

    Arguments:
        cell_doc (NeuroMLDocument): nml document of the cell model
        cell (ephys.CellModel): bluepyopt cell
        parameter (ephys.parameters)
        section_list (str): location
        included_channels (list): list of channels already included
            in the nml file
        skip_non_uniform (bool): True to skip non uniform distributions
        release_params (dict): optimized parameters
        skip_channels_copy (bool): True to skip the copy pasting
            of the neuroml channel files
        custom_channel_ion (dict): dict mapping channel to ion
        custom_ion_erevs (dict): dict mapping ion to erev (reversal potential)
    """
    channel = get_channel_from_param_name(parameter.param_name)

    cond_density, variable_parameters = extract_parameter_value(
        parameter, section_list, channel, skip_non_uniform, release_params
    )
    if cond_density is None and variable_parameters is None:
        return None, None

    channel_name = channel
    if channel in channel_substitutes:
        channel_name = channel_substitutes[channel]

    # add nml channel to nml cell file
    add_nml_channel_to_nml_cell_file(
        cell_doc,
        included_channels,
        channel_name=channel_name,
        skip_channels_copy=skip_channels_copy,
    )

    arguments, channel_class = get_arguments(
        params=cell.params,
        parameter_name=parameter.param_name,
        section_list=section_list,
        channel=channel,
        channel_name=channel_name,
        variable_parameters=variable_parameters,
        cond_density=cond_density,
        release_params=release_params,
        custom_channel_ion=custom_channel_ion,
        custom_ion_erevs=custom_ion_erevs,
    )

    density = getattr(neuroml, channel_class)(**arguments)

    return density, channel_class


def get_specific_capacitance(capacitance_overwrites):
    """Returns the specific capacitance.

    Arguments:
        capacitance_overwrites (dict): capacitance values from parameters
            to overwrites default ones.
    """
    specific_capacitances = []
    for section_list in default_capacitances:
        if section_list in capacitance_overwrites:
            capacitance = capacitance_overwrites[section_list]
        elif "all" in capacitance_overwrites:
            capacitance = capacitance_overwrites["all"]
        else:
            capacitance = default_capacitances[section_list]

        specific_capacitances.append(
            neuroml.SpecificCapacitance(
                value=capacitance, segment_groups=section_list
            )
        )

    return specific_capacitances


def get_biophys(
    cell,
    cell_doc,
    release_params,
    skip_non_uniform=False,
    skip_channels_copy=False,
    custom_channel_ion=None,
    custom_ion_erevs=None,
):
    """Get biophys in neuroml format.

    Arguments:
        cell (ephys.CellModel): bluepyopt cell
        cell_doc (NeuroMLDocument): nml document of the cell model
        release_params (dict): optimized parameters
        skip_non_uniform (bool): True to skip non uniform distributions
        skip_channels_copy (bool): True to skip the copy pasting
            of the neuroml channel files
        custom_channel_ion (dict): dict mapping channel to ion
        custom_ion_erevs (dict): dict mapping ion to erev (reversal potential)
    """
    concentrationModels = {}

    # Membrane properties
    included_channels = []
    channel_densities = []
    channel_density_nernsts = []
    channel_density_non_unif_nernsts = []
    channel_density_non_uniforms = []
    species = []

    capacitance_overwrites = {}

    for parameter in cell.params.values():
        if not (
            isinstance(parameter, ephys.parameters.NrnGlobalParameter)
            or isinstance(parameter, ephys.parameters.MetaParameter)
        ):
            for location in parameter.locations:
                section_list = location.seclist_name

                if (
                    parameter.param_name != "e_pas"
                    and "CaDynamics_E2" not in parameter.param_name
                    and parameter.param_name not in ignore_params
                ):
                    density, channel_class = get_density(
                        cell_doc,
                        cell,
                        parameter,
                        section_list,
                        included_channels,
                        skip_non_uniform,
                        release_params,
                        skip_channels_copy,
                        custom_channel_ion,
                        custom_ion_erevs,
                    )

                    if density is not None:
                        # add density to list of densities
                        if channel_class == "ChannelDensityNernst":
                            channel_density_nernsts.append(density)
                        elif (
                            channel_class == "ChannelDensityNernstNonUniform"
                        ):
                            channel_density_non_unif_nernsts.append(
                                density
                            )
                        elif channel_class == "ChannelDensityNonUniform":
                            channel_density_non_uniforms.append(density)
                        else:
                            channel_densities.append(density)

                elif "gamma_CaDynamics_E2" in parameter.param_name:
                    model = f"CaDynamics_E2_NML2__{cell.name}_{section_list}"
                    value = parameter.value
                    if value is None:
                        # non frozen parameter
                        value = release_params[parameter.name]

                    if model not in concentrationModels:
                        concentrationModels[model] = {}
                    concentrationModels[model]["gamma"] = value

                elif "decay_CaDynamics_E2" in parameter.param_name:
                    model = f"CaDynamics_E2_NML2__{cell.name}_{section_list}"
                    species.append(
                        neuroml.Species(
                            id="ca",
                            ion="ca",
                            initial_concentration="5.0E-11 mol_per_cm3",
                            initial_ext_concentration="2.0E-6 mol_per_cm3",
                            concentration_model=model,
                            segment_groups=section_list,
                        )
                    )

                    channel_nml2_file = "CaDynamics_E2_NML2.nml"
                    add_nml_channel_to_nml_cell_file(
                        cell_doc,
                        included_channels,
                        channel_nml2_file=channel_nml2_file,
                    )

                    value = parameter.value
                    if value is None:
                        # non frozen parameter
                        value = release_params[parameter.name]

                    if model not in concentrationModels:
                        concentrationModels[model] = {}
                    concentrationModels[model]["decay"] = value

                elif parameter.param_name == "cm":
                    capacitance_overwrites[
                        section_list
                    ] = f"{parameter.value} uF_per_cm2"

    # append new concentrations to the CaDynamics_E2_NML2.nml file
    if concentrationModels and not skip_channels_copy:
        adapt_CaDynamics_nml(concentrationModels)

    specific_capacitances = get_specific_capacitance(capacitance_overwrites)

    v_init = cell.params["v_init"].value
    init_memb_potentials = [
        neuroml.InitMembPotential(value=f"{v_init} mV", segment_groups="all")
    ]

    membrane_properties = neuroml.MembraneProperties(
        channel_densities=channel_densities,
        channel_density_nernsts=channel_density_nernsts,
        channel_density_non_uniform_nernsts=channel_density_non_unif_nernsts,
        channel_density_non_uniforms=channel_density_non_uniforms,
        specific_capacitances=specific_capacitances,
        init_memb_potentials=init_memb_potentials,
    )

    # Intracellular Properties
    Ra = cell.params["Ra.all"].value
    resistivities = [
        neuroml.Resistivity(value=f"{Ra} ohm_cm", segment_groups="all")
    ]

    intracellular_properties = neuroml.IntracellularProperties(
        resistivities=resistivities,
        species=species,
    )

    # Biophysical Properties
    biophysical_properties = neuroml.BiophysicalProperties(
        id="biophys",
        intracellular_properties=intracellular_properties,
        membrane_properties=membrane_properties,
    )

    return biophysical_properties



"""Optimisation class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0914, R0912


import random
import logging
import shutil
import os

import deap.algorithms
import deap.tools
import pickle

from .stoppingCriteria import MaxNGen
from . import utils

logger = logging.getLogger('__main__')


def _define_fitness(pop, obj_size):
    ''' Re-instanciate the fitness of the individuals for it to matches the
    evaluation function.
    '''
    from .optimisations import WSListIndividual

    new_pop = []
    if pop:
        for ind in pop:
            new_pop.append(WSListIndividual(list(ind), obj_size=obj_size))

    return new_pop


def _evaluate_invalid_fitness(toolbox, population):
    '''Evaluate the individuals with an invalid fitness

    Returns the count of individuals with invalid fitness
    '''
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    return len(invalid_ind)


def _get_offspring(parents, toolbox, cxpb, mutpb):
    '''return the offspring, use toolbox.variate if possible'''
    if hasattr(toolbox, 'variate'):
        return toolbox.variate(parents, toolbox, cxpb, mutpb)
    return deap.algorithms.varAnd(parents, toolbox, cxpb, mutpb)


def _check_stopping_criteria(criteria, params):
    for c in criteria:
        c.check(params)
        if c.criteria_met:
            logger.info('Run stopped because of stopping criteria: ' +
                        c.name)
            return True
    else:
        return False


def eaAlphaMuPlusLambdaCheckpoint(
        population,
        toolbox,
        mu,
        cxpb,
        mutpb,
        ngen,
        stats=None,
        halloffame=None,
        cp_frequency=1,
        cp_filename=None,
        continue_cp=False,
        terminator=None,
        param_names=None):
    r"""This is the :math:`(~\alpha,\mu~,~\lambda)` evolutionary algorithm

    Args:
        population(list of deap Individuals)
        toolbox(deap Toolbox)
        mu(int): Total parent population size of EA
        cxpb(float): Crossover probability
        mutpb(float): Mutation probability
        ngen(int): Total number of generation to run
        stats(deap.tools.Statistics): generation of statistics
        halloffame(deap.tools.HallOfFame): hall of fame
        cp_frequency(int): generations between checkpoints
        cp_filename(string): path to checkpoint filename
        continue_cp(bool): whether to continue
        terminator (multiprocessing.Event): exit loop when is set.
            Not taken into account if None.
        param_names(list): names of the parameters optimized by the evaluator
    """

    if param_names is None:
        param_names = []

    if cp_filename:
        cp_filename_tmp = cp_filename + '.tmp'

    if continue_cp:
        # A file name has been given, then load the data from the file
        with open(cp_filename, "rb") as f:
            cp = pickle.load(f)
        population = cp["population"]
        parents = cp["parents"]
        start_gen = cp["generation"]
        halloffame = cp["halloffame"]
        logbook = cp["logbook"]
        history = cp["history"]
        random.setstate(cp["rndstate"])

        # Assert that the fitness of the individuals match the evaluator
        obj_size = len(population[0].fitness.wvalues)
        population = _define_fitness(population, obj_size)
        parents = _define_fitness(parents, obj_size)
        _evaluate_invalid_fitness(toolbox, parents)
        _evaluate_invalid_fitness(toolbox, population)

    else:
        # Start a new evolution
        start_gen = 1
        parents = population[:]
        logbook = deap.tools.Logbook()
        logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])
        history = deap.tools.History()

        invalid_count = _evaluate_invalid_fitness(toolbox, population)

        utils.update_history_and_hof(halloffame, history, population)
        utils.record_stats(
            stats, logbook, start_gen, population, invalid_count
        )

    stopping_criteria = [MaxNGen(ngen)]

    # Begin the generational process
    gen = start_gen + 1
    stopping_params = {"gen": gen}
    while utils.run_next_gen(
            not (_check_stopping_criteria(stopping_criteria, stopping_params)),
            terminator):
        offspring = _get_offspring(parents, toolbox, cxpb, mutpb)

        population = parents + offspring

        invalid_count = _evaluate_invalid_fitness(toolbox, offspring)
        utils.update_history_and_hof(halloffame, history, population)
        utils.record_stats(stats, logbook, gen, population, invalid_count)

        # Select the next generation parents
        parents = toolbox.select(population, mu)

        logger.info(logbook.stream)

        if (cp_filename and cp_frequency and
           gen % cp_frequency == 0):
            cp = dict(population=population,
                      generation=gen,
                      parents=parents,
                      halloffame=halloffame,
                      history=history,
                      logbook=logbook,
                      rndstate=random.getstate(),
                      param_names=param_names)
            with open(cp_filename_tmp, "wb") as f:
                pickle.dump(cp, f)
            if os.path.isfile(cp_filename_tmp):
                shutil.copy(cp_filename_tmp, cp_filename)
                logger.debug('Wrote checkpoint to %s', cp_filename)

        gen += 1
        stopping_params["gen"] = gen

    return population, halloffame, logbook, history



"""CMA Optimisation class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import logging
import numpy
import pickle
import random
import functools
import shutil
import os

import deap.tools

from .CMA_SO import CMA_SO
from .CMA_MO import CMA_MO
from . import utils

import bluepyopt.optimisations

logger = logging.getLogger("__main__")


def _ind_convert_space(ind, convert_fcn):
    """util function to pass the individual from normalized to real space and
    inversely"""
    return [f(x) for f, x in zip(convert_fcn, ind)]


class DEAPOptimisationCMA(bluepyopt.optimisations.Optimisation):

    """Optimisation class for CMA-based evolution strategies"""

    def __init__(
        self,
        evaluator=None,
        use_scoop=False,
        seed=1,
        offspring_size=None,
        centroids=None,
        sigma=0.4,
        map_function=None,
        hof=None,
        selector_name="single_objective",
        weight_hv=0.5,
        fitness_reduce=numpy.sum,
        use_stagnation_criterion=True,
    ):
        """Constructor

        Args:
            evaluator (Evaluator): Evaluator object
            use_scoop (bool): use scoop map for parallel computation
            seed (float): Random number generator seed
            offspring_size (int): Number of offspring individuals in each
                generation
            centroids (list): list of initial guesses used as the starting
                points of the CMA-ES
            sigma (float): initial standard deviation of the distribution
            map_function (function): Function used to map (parallelize) the
                evaluation function calls
            hof (hof): Hall of Fame object
            selector_name (str): The selector used in the evolutionary
                algorithm, possible values are 'single_objective' or
                'multi_objective'
            weight_hv (float): between 0 and 1. Weight given to the
                hyper-volume contribution when computing the score of an
                individual in MO-CMA. The weight of the fitness contribution
                is computed as 1 - weight_hv.
            fitness_reduce (fcn): function used to reduce the objective values
                to a single fitness score
            use_stagnation_criterion (bool): whether to use the stagnation
                stopping criterion on top of the maximum generation criterion
        """

        super(DEAPOptimisationCMA, self).__init__(evaluator=evaluator)

        self.use_scoop = use_scoop
        self.seed = seed
        self.map_function = map_function

        self.hof = hof
        if self.hof is None:
            self.hof = deap.tools.HallOfFame(10)

        self.offspring_size = offspring_size

        self.fitness_reduce = fitness_reduce
        self.centroids = centroids
        self.sigma = sigma

        if weight_hv > 1.0 or weight_hv < 0.0:
            raise Exception("weight_hv has to be between 0 and 1.")
        self.weight_hv = weight_hv

        self.selector_name = selector_name
        if self.selector_name == "single_objective":
            self.cma_creator = CMA_SO
        elif self.selector_name == "multi_objective":
            self.cma_creator = CMA_MO
        else:
            raise Exception(
                "The selector_name has to be 'single_objective' "
                "or 'multi_objective'. Not "
                "{}".format(self.selector_name)
            )

        self.use_stagnation_criterion = use_stagnation_criterion

        # Number of objective values
        self.problem_size = len(self.evaluator.params)

        # Number of parameters
        self.ind_size = len(self.evaluator.objectives)

        # Create a DEAP toolbox
        self.toolbox = deap.base.Toolbox()

        # Bounds for the parameters
        self.lbounds = [p.lower_bound for p in self.evaluator.params]
        self.ubounds = [p.upper_bound for p in self.evaluator.params]

        # Instantiate functions converting individuals from the original
        # parameter space to (and from) a normalized space bounded to [-1.;1]
        self.ubounds = numpy.asarray(self.ubounds)
        self.lbounds = numpy.asarray(self.lbounds)
        bounds_radius = (self.ubounds - self.lbounds) / 2.0
        bounds_mean = (self.ubounds + self.lbounds) / 2.0

        self.to_norm = []
        self.to_space = []
        for r, m in zip(bounds_radius, bounds_mean):
            self.to_norm.append(
                functools.partial(
                    lambda param, bm, br: (param - bm) / br,
                    bm=m,
                    br=r)
            )
            self.to_space.append(
                functools.partial(
                    lambda param, bm, br: (param * br) + bm,
                    bm=m,
                    br=r
                )
            )

        # Overwrite the bounds with -1. and 1.
        self.lbounds = numpy.full(self.problem_size, -1.0)
        self.ubounds = numpy.full(self.problem_size, 1.0)

        self.setup_deap()

        # In case initial guesses were provided, rescale them to the norm space
        if self.centroids is not None:
            self.centroids = [
                self.toolbox.Individual(_ind_convert_space(ind, self.to_norm))
                for ind in centroids
            ]

    def setup_deap(self):
        """Set up optimisation"""

        # Set random seed
        random.seed(self.seed)
        numpy.random.seed(self.seed)

        # Register the 'uniform' function
        self.toolbox.register(
            "uniformparams",
            utils.uniform,
            self.lbounds,
            self.ubounds,
            self.ind_size
        )

        # Register the individual format
        self.toolbox.register(
            "Individual",
            functools.partial(
                utils.WSListIndividual,
                obj_size=self.ind_size,
                reduce_fcn=self.fitness_reduce,
            ),
        )

        # A Random Individual is created by ListIndividual and parameters are
        # initially picked by 'uniform'
        self.toolbox.register(
            "RandomInd",
            deap.tools.initIterate,
            self.toolbox.Individual,
            self.toolbox.uniformparams,
        )

        # Register the population format. It is a list of individuals
        self.toolbox.register(
            "population", deap.tools.initRepeat, list, self.toolbox.RandomInd
        )

        # Register the evaluation function for the individuals
        self.toolbox.register(
            "evaluate",
            self.evaluator.init_simulator_and_evaluate_with_lists
        )

        import copyreg
        import types

        copyreg.pickle(types.MethodType, utils.reduce_method)

        if self.use_scoop:
            if self.map_function:
                raise Exception(
                    "Impossible to use scoop is providing self defined map "
                    "function: %s" % self.map_function
                )

            from scoop import futures

            self.toolbox.register("map", futures.map)

        elif self.map_function:
            self.toolbox.register("map", self.map_function)

    def run(
        self,
        max_ngen=0,
        cp_frequency=1,
        continue_cp=False,
        cp_filename=None,
        terminator=None,
    ):
        """ Run the optimizer until a stopping criteria is met.

        Args:
            max_ngen(int): Total number of generation to run
            cp_frequency(int): generations between checkpoints
            continue_cp(bool): whether to continue
            cp_filename(string): path to checkpoint filename
            terminator (multiprocessing.Event): exit loop when is set.
                Not taken into account if None.
        """
        if cp_filename:
            cp_filename_tmp = cp_filename + '.tmp'

        stats = self.get_stats()

        if continue_cp:

            # A file name has been given, then load the data from the file
            with open(cp_filename, "rb") as f:
                cp = pickle.load(f)
            gen = cp["generation"]
            self.hof = cp["halloffame"]
            logbook = cp["logbook"]
            history = cp["history"]
            random.setstate(cp["rndstate"])
            numpy.random.set_state(cp["np_rndstate"])
            CMA_es = cp["CMA_es"]
            CMA_es.map_function = self.map_function

        else:

            history = deap.tools.History()
            logbook = deap.tools.Logbook()
            logbook.header = ["gen", "nevals"] + stats.fields

            # Instantiate the CMA strategy centered on the centroids
            CMA_es = self.cma_creator(
                centroids=self.centroids,
                offspring_size=self.offspring_size,
                sigma=self.sigma,
                max_ngen=max_ngen,
                IndCreator=self.toolbox.Individual,
                RandIndCreator=self.toolbox.RandomInd,
                map_function=self.map_function,
                use_scoop=self.use_scoop,
                use_stagnation_criterion=self.use_stagnation_criterion,
            )

            if self.selector_name == "multi_objective":
                CMA_es.weight_hv = self.weight_hv
                to_evaluate = CMA_es.get_parents(self.to_space)
                fitness = self.toolbox.map(self.toolbox.evaluate, to_evaluate)
                fitness = list(map(list, fitness))
                CMA_es.set_fitness_parents(fitness)

            gen = 1

        pop = CMA_es.get_population(self.to_space)

        param_names = []
        if hasattr(self.evaluator, "param_names"):
            param_names = self.evaluator.param_names

        # Run until a termination criteria is met
        while utils.run_next_gen(CMA_es.active, terminator):
            logger.info("Generation {}".format(gen))

            # Generate the new populations
            n_out = CMA_es.generate_new_pop(
                lbounds=self.lbounds, ubounds=self.ubounds
            )
            logger.debug(
                "Number of individuals outside of bounds: {} ({:.2f}%)".format(
                    n_out,
                    100.0 * n_out / len(CMA_es.population)
                )
            )

            # Get all the individuals in the original space for evaluation
            to_evaluate = CMA_es.get_population(self.to_space)

            # Compute the fitness
            fitness = self.toolbox.map(self.toolbox.evaluate, to_evaluate)
            fitness = list(map(list, fitness))
            nevals = len(to_evaluate)
            CMA_es.set_fitness(fitness)

            # Update the hall of fame, history and logbook
            pop = CMA_es.get_population(self.to_space)
            utils.update_history_and_hof(self.hof, history, pop)
            record = utils.record_stats(stats, logbook, gen, pop, nevals)
            logger.info(logbook.stream)

            # Update the CMA strategy using the new fitness and check if
            # termination conditions were reached
            CMA_es.update_strategy()
            CMA_es.check_termination(gen)

            if cp_filename and cp_frequency and gen % cp_frequency == 0:

                # Map function shouldn't be pickled
                temp_mf = CMA_es.map_function
                CMA_es.map_function = None

                cp = dict(
                    population=pop,
                    generation=gen,
                    halloffame=self.hof,
                    history=history,
                    logbook=logbook,
                    rndstate=random.getstate(),
                    np_rndstate=numpy.random.get_state(),
                    CMA_es=CMA_es,
                    param_names=param_names,
                )
                with open(cp_filename_tmp, "wb") as f:
                    pickle.dump(cp, f)
                if os.path.isfile(cp_filename_tmp):
                    shutil.copy(cp_filename_tmp, cp_filename)
                    logger.debug("Wrote checkpoint to %s", cp_filename)

                CMA_es.map_function = temp_mf

            gen += 1

        return pop, self.hof, logbook, history

    def get_stats(self):
        """Get the stats that will be saved during optimisation"""
        stats = deap.tools.Statistics(key=lambda ind: ind.fitness.reduce)
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)
        return stats



"""ephys/morphologies.py unit tests"""

import json
import os


import pytest

import bluepyopt.ephys as ephys
from bluepyopt.ephys.serializer import instantiator

testdata_dir = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)),
    'testdata')

simpleswc_morphpath = os.path.join(testdata_dir, 'simple.swc')
simpleswc_ax1_morphpath = os.path.join(testdata_dir, 'simple_ax1.swc')
simpleswc_ax2_morphpath = os.path.join(testdata_dir, 'simple_ax2.asc')
simplewrong_morphpath = os.path.join(testdata_dir, 'simple.wrong')


@pytest.mark.unit
def test_morphology_init():
    """ephys.morphologies: testing Morphology constructor"""

    morph = ephys.morphologies.Morphology()
    assert isinstance(morph, ephys.morphologies.Morphology)


@pytest.mark.unit
def test_nrnfilemorphology_init():
    """ephys.morphologies: testing NrnFileMorphology constructor"""
    sim = ephys.simulators.NrnSimulator()

    morph = ephys.morphologies.NrnFileMorphology('wrong.swc')
    pytest.raises(IOError, morph.instantiate, sim=sim)

    morph = ephys.morphologies.NrnFileMorphology(simplewrong_morphpath)

    assert str(morph) == simplewrong_morphpath

    pytest.raises(ValueError, morph.instantiate, sim=sim)

    morph = ephys.morphologies.NrnFileMorphology(
        simpleswc_morphpath,
        do_set_nseg=False)

    morph.instantiate(sim=sim)
    morph.destroy(sim=sim)


@pytest.mark.unit
def test_nrnfilemorphology_replace_axon():
    """ephys.morphologies: testing NrnFileMorphology replace_axon"""
    sim = ephys.simulators.NrnSimulator()

    morph = ephys.morphologies.NrnFileMorphology(
        simpleswc_morphpath,
        do_replace_axon=True)

    cell = ephys.models.CellModel(name='cell_replace')
    icell = cell.create_empty_cell(
        cell.name,
        sim=sim,
        seclist_names=cell.seclist_names,
        secarray_names=cell.secarray_names)

    morph.instantiate(sim=sim, icell=icell)

    assert len([sec for sec in icell.axon]) == 2

    morph.destroy(sim=sim)
    icell.destroy()


@pytest.mark.unit
def test_nrnfilemorphology_replace_axon_ax1():
    """ephys.morphologies: testing NrnFileMorphology replace_axon with ax1"""
    sim = ephys.simulators.NrnSimulator()

    morph = ephys.morphologies.NrnFileMorphology(
        simpleswc_ax1_morphpath,
        do_replace_axon=True)

    cell = ephys.models.CellModel(name='cell_ax1')
    icell = cell.create_empty_cell(
        cell.name,
        sim=sim,
        seclist_names=cell.seclist_names,
        secarray_names=cell.secarray_names)

    morph.instantiate(sim=sim, icell=icell)

    assert len([sec for sec in icell.axon]) == 2

    morph.destroy(sim=sim)
    icell.destroy()


@pytest.mark.unit
def test_nrnfilemorphology_replace_axon_ax2():
    """ephys.morphologies: testing NrnFileMorphology replace_axon with ax2"""
    sim = ephys.simulators.NrnSimulator()

    morph = ephys.morphologies.NrnFileMorphology(
        simpleswc_ax2_morphpath,
        do_replace_axon=True)

    cell = ephys.models.CellModel(name='cell_ax2')
    icell = cell.create_empty_cell(
        cell.name,
        sim=sim,
        seclist_names=cell.seclist_names,
        secarray_names=cell.secarray_names)

    morph.instantiate(sim=sim, icell=icell)

    assert len([sec for sec in icell.axon]) == 2

    morph.destroy(sim=sim)
    icell.destroy()


@pytest.mark.unit
def test_serialize():
    """ephys.morphology: testing serialization"""

    morph = ephys.morphologies.NrnFileMorphology(simpleswc_morphpath)
    serialized = morph.to_dict()
    assert isinstance(json.dumps(serialized), str)
    deserialized = instantiator(serialized)
    assert isinstance(deserialized, ephys.morphologies.NrnFileMorphology)
    assert deserialized.morphology_path == simpleswc_morphpath
